# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Id3v24(KaitaiStruct):
    """
    .. seealso::
       Source - http://id3.org/id3v2.4.0-structure
    
    
    .. seealso::
       Source - http://id3.org/id3v2.4.0-frames
    """
    SEQ_FIELDS = ["tag"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['tag']['start'] = self._io.pos()
        self.tag = Id3v24.Tag(self._io, self, self._root)
        self.tag._read()
        self._debug['tag']['end'] = self._io.pos()

    class U1beSynchsafe(KaitaiStruct):
        SEQ_FIELDS = ["padding", "value"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['padding']['start'] = self._io.pos()
            self.padding = self._io.read_bits_int_be(1) != 0
            self._debug['padding']['end'] = self._io.pos()
            self._debug['value']['start'] = self._io.pos()
            self.value = self._io.read_bits_int_be(7)
            self._debug['value']['end'] = self._io.pos()


    class U2beSynchsafe(KaitaiStruct):
        SEQ_FIELDS = ["byte0", "byte1"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['byte0']['start'] = self._io.pos()
            self.byte0 = Id3v24.U1beSynchsafe(self._io, self, self._root)
            self.byte0._read()
            self._debug['byte0']['end'] = self._io.pos()
            self._debug['byte1']['start'] = self._io.pos()
            self.byte1 = Id3v24.U1beSynchsafe(self._io, self, self._root)
            self.byte1._read()
            self._debug['byte1']['end'] = self._io.pos()

        @property
        def value(self):
            if hasattr(self, '_m_value'):
                return self._m_value if hasattr(self, '_m_value') else None

            self._m_value = ((self.byte0.value << 7) | self.byte1.value)
            return self._m_value if hasattr(self, '_m_value') else None


    class Tag(KaitaiStruct):
        SEQ_FIELDS = ["header", "header_ex", "frames", "padding", "footer"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['header']['start'] = self._io.pos()
            self.header = Id3v24.Header(self._io, self, self._root)
            self.header._read()
            self._debug['header']['end'] = self._io.pos()
            if self.header.flags.flag_headerex:
                self._debug['header_ex']['start'] = self._io.pos()
                self.header_ex = Id3v24.HeaderEx(self._io, self, self._root)
                self.header_ex._read()
                self._debug['header_ex']['end'] = self._io.pos()

            self._debug['frames']['start'] = self._io.pos()
            self.frames = []
            i = 0
            while True:
                if not 'arr' in self._debug['frames']:
                    self._debug['frames']['arr'] = []
                self._debug['frames']['arr'].append({'start': self._io.pos()})
                _t_frames = Id3v24.Frame(self._io, self, self._root)
                _t_frames._read()
                _ = _t_frames
                self.frames.append(_)
                self._debug['frames']['arr'][len(self.frames) - 1]['end'] = self._io.pos()
                if  (((self._io.pos() + _.size.value) > self.header.size.value) or (_.is_invalid)) :
                    break
                i += 1
            self._debug['frames']['end'] = self._io.pos()
            if not (self.header.flags.flag_footer):
                self._debug['padding']['start'] = self._io.pos()
                self.padding = Id3v24.Padding(self._io, self, self._root)
                self.padding._read()
                self._debug['padding']['end'] = self._io.pos()

            if self.header.flags.flag_footer:
                self._debug['footer']['start'] = self._io.pos()
                self.footer = Id3v24.Footer(self._io, self, self._root)
                self.footer._read()
                self._debug['footer']['end'] = self._io.pos()



    class U4beSynchsafe(KaitaiStruct):
        SEQ_FIELDS = ["short0", "short1"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['short0']['start'] = self._io.pos()
            self.short0 = Id3v24.U2beSynchsafe(self._io, self, self._root)
            self.short0._read()
            self._debug['short0']['end'] = self._io.pos()
            self._debug['short1']['start'] = self._io.pos()
            self.short1 = Id3v24.U2beSynchsafe(self._io, self, self._root)
            self.short1._read()
            self._debug['short1']['end'] = self._io.pos()

        @property
        def value(self):
            if hasattr(self, '_m_value'):
                return self._m_value if hasattr(self, '_m_value') else None

            self._m_value = ((self.short0.value << 14) | self.short1.value)
            return self._m_value if hasattr(self, '_m_value') else None


    class Frame(KaitaiStruct):
        SEQ_FIELDS = ["id", "size", "flags_status", "flags_format", "data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['id']['start'] = self._io.pos()
            self.id = (self._io.read_bytes(4)).decode(u"ASCII")
            self._debug['id']['end'] = self._io.pos()
            self._debug['size']['start'] = self._io.pos()
            self.size = Id3v24.U4beSynchsafe(self._io, self, self._root)
            self.size._read()
            self._debug['size']['end'] = self._io.pos()
            self._debug['flags_status']['start'] = self._io.pos()
            self.flags_status = Id3v24.Frame.FlagsStatus(self._io, self, self._root)
            self.flags_status._read()
            self._debug['flags_status']['end'] = self._io.pos()
            self._debug['flags_format']['start'] = self._io.pos()
            self.flags_format = Id3v24.Frame.FlagsFormat(self._io, self, self._root)
            self.flags_format._read()
            self._debug['flags_format']['end'] = self._io.pos()
            self._debug['data']['start'] = self._io.pos()
            self.data = self._io.read_bytes(self.size.value)
            self._debug['data']['end'] = self._io.pos()

        class FlagsStatus(KaitaiStruct):
            SEQ_FIELDS = ["reserved1", "flag_discard_alter_tag", "flag_discard_alter_file", "flag_read_only", "reserved2"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved1']['start'] = self._io.pos()
                self.reserved1 = self._io.read_bits_int_be(1) != 0
                self._debug['reserved1']['end'] = self._io.pos()
                self._debug['flag_discard_alter_tag']['start'] = self._io.pos()
                self.flag_discard_alter_tag = self._io.read_bits_int_be(1) != 0
                self._debug['flag_discard_alter_tag']['end'] = self._io.pos()
                self._debug['flag_discard_alter_file']['start'] = self._io.pos()
                self.flag_discard_alter_file = self._io.read_bits_int_be(1) != 0
                self._debug['flag_discard_alter_file']['end'] = self._io.pos()
                self._debug['flag_read_only']['start'] = self._io.pos()
                self.flag_read_only = self._io.read_bits_int_be(1) != 0
                self._debug['flag_read_only']['end'] = self._io.pos()
                self._debug['reserved2']['start'] = self._io.pos()
                self.reserved2 = self._io.read_bits_int_be(4)
                self._debug['reserved2']['end'] = self._io.pos()


        class FlagsFormat(KaitaiStruct):
            SEQ_FIELDS = ["reserved1", "flag_grouping", "reserved2", "flag_compressed", "flag_encrypted", "flag_unsynchronisated", "flag_indicator"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved1']['start'] = self._io.pos()
                self.reserved1 = self._io.read_bits_int_be(1) != 0
                self._debug['reserved1']['end'] = self._io.pos()
                self._debug['flag_grouping']['start'] = self._io.pos()
                self.flag_grouping = self._io.read_bits_int_be(1) != 0
                self._debug['flag_grouping']['end'] = self._io.pos()
                self._debug['reserved2']['start'] = self._io.pos()
                self.reserved2 = self._io.read_bits_int_be(2)
                self._debug['reserved2']['end'] = self._io.pos()
                self._debug['flag_compressed']['start'] = self._io.pos()
                self.flag_compressed = self._io.read_bits_int_be(1) != 0
                self._debug['flag_compressed']['end'] = self._io.pos()
                self._debug['flag_encrypted']['start'] = self._io.pos()
                self.flag_encrypted = self._io.read_bits_int_be(1) != 0
                self._debug['flag_encrypted']['end'] = self._io.pos()
                self._debug['flag_unsynchronisated']['start'] = self._io.pos()
                self.flag_unsynchronisated = self._io.read_bits_int_be(1) != 0
                self._debug['flag_unsynchronisated']['end'] = self._io.pos()
                self._debug['flag_indicator']['start'] = self._io.pos()
                self.flag_indicator = self._io.read_bits_int_be(1) != 0
                self._debug['flag_indicator']['end'] = self._io.pos()


        @property
        def is_invalid(self):
            if hasattr(self, '_m_is_invalid'):
                return self._m_is_invalid if hasattr(self, '_m_is_invalid') else None

            self._m_is_invalid = self.id == u"\000\000\000\000"
            return self._m_is_invalid if hasattr(self, '_m_is_invalid') else None


    class HeaderEx(KaitaiStruct):
        SEQ_FIELDS = ["size", "flags_ex", "data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['size']['start'] = self._io.pos()
            self.size = Id3v24.U4beSynchsafe(self._io, self, self._root)
            self.size._read()
            self._debug['size']['end'] = self._io.pos()
            self._debug['flags_ex']['start'] = self._io.pos()
            self.flags_ex = Id3v24.HeaderEx.FlagsEx(self._io, self, self._root)
            self.flags_ex._read()
            self._debug['flags_ex']['end'] = self._io.pos()
            self._debug['data']['start'] = self._io.pos()
            self.data = self._io.read_bytes((self.size.value - 5))
            self._debug['data']['end'] = self._io.pos()

        class FlagsEx(KaitaiStruct):
            SEQ_FIELDS = ["reserved1", "flag_update", "flag_crc", "flag_restrictions", "reserved2"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved1']['start'] = self._io.pos()
                self.reserved1 = self._io.read_bits_int_be(1) != 0
                self._debug['reserved1']['end'] = self._io.pos()
                self._debug['flag_update']['start'] = self._io.pos()
                self.flag_update = self._io.read_bits_int_be(1) != 0
                self._debug['flag_update']['end'] = self._io.pos()
                self._debug['flag_crc']['start'] = self._io.pos()
                self.flag_crc = self._io.read_bits_int_be(1) != 0
                self._debug['flag_crc']['end'] = self._io.pos()
                self._debug['flag_restrictions']['start'] = self._io.pos()
                self.flag_restrictions = self._io.read_bits_int_be(1) != 0
                self._debug['flag_restrictions']['end'] = self._io.pos()
                self._debug['reserved2']['start'] = self._io.pos()
                self.reserved2 = self._io.read_bits_int_be(4)
                self._debug['reserved2']['end'] = self._io.pos()



    class Header(KaitaiStruct):
        SEQ_FIELDS = ["magic", "version_major", "version_revision", "flags", "size"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(3)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x49\x44\x33":
                raise kaitaistruct.ValidationNotEqualError(b"\x49\x44\x33", self.magic, self._io, u"/types/header/seq/0")
            self._debug['version_major']['start'] = self._io.pos()
            self.version_major = self._io.read_u1()
            self._debug['version_major']['end'] = self._io.pos()
            self._debug['version_revision']['start'] = self._io.pos()
            self.version_revision = self._io.read_u1()
            self._debug['version_revision']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = Id3v24.Header.Flags(self._io, self, self._root)
            self.flags._read()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['size']['start'] = self._io.pos()
            self.size = Id3v24.U4beSynchsafe(self._io, self, self._root)
            self.size._read()
            self._debug['size']['end'] = self._io.pos()

        class Flags(KaitaiStruct):
            SEQ_FIELDS = ["flag_unsynchronization", "flag_headerex", "flag_experimental", "flag_footer", "reserved"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['flag_unsynchronization']['start'] = self._io.pos()
                self.flag_unsynchronization = self._io.read_bits_int_be(1) != 0
                self._debug['flag_unsynchronization']['end'] = self._io.pos()
                self._debug['flag_headerex']['start'] = self._io.pos()
                self.flag_headerex = self._io.read_bits_int_be(1) != 0
                self._debug['flag_headerex']['end'] = self._io.pos()
                self._debug['flag_experimental']['start'] = self._io.pos()
                self.flag_experimental = self._io.read_bits_int_be(1) != 0
                self._debug['flag_experimental']['end'] = self._io.pos()
                self._debug['flag_footer']['start'] = self._io.pos()
                self.flag_footer = self._io.read_bits_int_be(1) != 0
                self._debug['flag_footer']['end'] = self._io.pos()
                self._debug['reserved']['start'] = self._io.pos()
                self.reserved = self._io.read_bits_int_be(4)
                self._debug['reserved']['end'] = self._io.pos()



    class Padding(KaitaiStruct):
        SEQ_FIELDS = ["padding"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['padding']['start'] = self._io.pos()
            self.padding = self._io.read_bytes((self._root.tag.header.size.value - self._io.pos()))
            self._debug['padding']['end'] = self._io.pos()


    class Footer(KaitaiStruct):
        SEQ_FIELDS = ["magic", "version_major", "version_revision", "flags", "size"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(3)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x33\x44\x49":
                raise kaitaistruct.ValidationNotEqualError(b"\x33\x44\x49", self.magic, self._io, u"/types/footer/seq/0")
            self._debug['version_major']['start'] = self._io.pos()
            self.version_major = self._io.read_u1()
            self._debug['version_major']['end'] = self._io.pos()
            self._debug['version_revision']['start'] = self._io.pos()
            self.version_revision = self._io.read_u1()
            self._debug['version_revision']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = Id3v24.Footer.Flags(self._io, self, self._root)
            self.flags._read()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['size']['start'] = self._io.pos()
            self.size = Id3v24.U4beSynchsafe(self._io, self, self._root)
            self.size._read()
            self._debug['size']['end'] = self._io.pos()

        class Flags(KaitaiStruct):
            SEQ_FIELDS = ["flag_unsynchronization", "flag_headerex", "flag_experimental", "flag_footer", "reserved"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['flag_unsynchronization']['start'] = self._io.pos()
                self.flag_unsynchronization = self._io.read_bits_int_be(1) != 0
                self._debug['flag_unsynchronization']['end'] = self._io.pos()
                self._debug['flag_headerex']['start'] = self._io.pos()
                self.flag_headerex = self._io.read_bits_int_be(1) != 0
                self._debug['flag_headerex']['end'] = self._io.pos()
                self._debug['flag_experimental']['start'] = self._io.pos()
                self.flag_experimental = self._io.read_bits_int_be(1) != 0
                self._debug['flag_experimental']['end'] = self._io.pos()
                self._debug['flag_footer']['start'] = self._io.pos()
                self.flag_footer = self._io.read_bits_int_be(1) != 0
                self._debug['flag_footer']['end'] = self._io.pos()
                self._debug['reserved']['start'] = self._io.pos()
                self.reserved = self._io.read_bits_int_be(4)
                self._debug['reserved']['end'] = self._io.pos()







import mmap
import os
from pathlib import Path
import tempfile as tf
import sys
from typing import IO, Optional, Union


def make_stream(path_or_stream, mode='rb', close_on_exit=None):
    if isinstance(path_or_stream, FileStream):
        return path_or_stream
    else:
        return FileStream(path_or_stream, mode=mode, close_on_exit=close_on_exit)


class Tempfile:
    def __init__(self, contents, prefix=None, suffix=None):
        self._temp = None
        self._data = contents
        self._prefix = prefix
        self._suffix = suffix

    def __enter__(self):
        self._temp = tf.NamedTemporaryFile(prefix=self._prefix, suffix=self._suffix, delete=False)
        self._temp.write(self._data)
        self._temp.flush()
        self._temp.close()
        return self._temp.name

    def __exit__(self, type, value, traceback):
        if self._temp is not None:
            os.unlink(self._temp.name)
            self._temp = None


class PathOrStdin:
    def __init__(self, path):
        self._path = path
        if self._path == '-':
            self._tempfile = Tempfile(sys.stdin.buffer.read())
        else:
            self._tempfile = None

    def __enter__(self):
        if self._tempfile is None:
            return self._path
        else:
            return self._tempfile.__enter__()

    def __exit__(self, *args, **kwargs):
        if self._tempfile is not None:
            return self._tempfile.__exit__(*args, **kwargs)


class FileStream:
    def __init__(
            self,
            path_or_stream: Union[str, Path, IO, "FileStream"],
            start: int = 0,
            length: Optional[int] = None,
            mode: str = "rb",
            close_on_exit: Optional[bool] = None
    ):
        if isinstance(path_or_stream, Path):
            path_or_stream = str(path_or_stream)
        if isinstance(path_or_stream, str):
            self._stream: IO = open(path_or_stream, mode)
            if close_on_exit is None:
                close_on_exit = True
        else:
            if not path_or_stream.seekable():
                raise ValueError('FileStream can only wrap streams that are seekable')
            elif not path_or_stream.readable():
                raise ValueError('FileStream can only wrap streams that are readable')
            self._stream = path_or_stream
        if isinstance(path_or_stream, FileStream):
            if length is None:
                self._length = len(path_or_stream) - start
            else:
                self._length = min(length, len(path_or_stream))
        else:
            filesize = os.path.getsize(self._stream.name)
            if length is None:
                self._length = filesize - start
            else:
                self._length = min(filesize, length) - start
        if close_on_exit is None:
            close_on_exit = False
        self._name = self._stream.name
        self.start = start
        self.close_on_exit = close_on_exit
        self._entries = 0
        self._listeners = []
        self._root = None

    def __len__(self):
        return self._length

    def add_listener(self, listener):
        self._listeners.append(listener)

    def remove_listener(self, listener):
        ret = False
        for i in reversed(range(len(self._listeners))):
            if self._listeners[i] == listener:
                del self._listeners[i]
                ret = True
        return ret

    def seekable(self):
        return True

    def writable(self):
        return False

    def readable(self):
        return True

    @property
    def name(self):
        return self._name

    @property
    def root(self):
        if self._root is None:
            if isinstance(self._stream, FileStream):
                self._root = self._stream.root
            else:
                self._root = self._stream
        return self._root

    def save_pos(self):
        f = self

        class SP:
            def __init__(self):
                self.pos = f.root.tell()

            def __enter__(self, *args, **kwargs):
                return f

            def __exit__(self, *args, **kwargs):
                f.root.seek(self.pos)

        return SP()

    def fileno(self):
        return self._stream.fileno()

    def offset(self):
        if isinstance(self._stream, FileStream):
            return self._stream.offset() + self.start
        else:
            return self.start

    def seek(self, offset, from_what=0):
        if from_what == 1:
            offset = self.tell() + offset
        elif from_what == 2:
            offset = len(self) + offset
        if offset - self.start > self._length:
            raise IndexError(f"{self!r} is {len(self)} bytes long, but seek was requested for byte {offset}")
        self._stream.seek(self.start + offset)

    def tell(self):
        return min(max(self._stream.tell() - self.start, 0), self._length)

    def read(self, n=None, update_listeners=True):
        if self._stream.tell() - self.start < 0:
            # another context moved the position, so move it back to our zero index:
            self.seek(0)
            pos = 0
        else:
            pos = self.tell()
        if update_listeners:
            for listener in self._listeners:
                listener(self, pos)
        ls = len(self)
        if pos >= ls:
            return b''
        elif n is None:
            return self._stream.read()[:ls - pos]
        else:
            return self._stream.read(min(n, ls - pos))

    def contains_all(self, *args):
        if args:
            with mmap.mmap(self.fileno(), 0, access=mmap.ACCESS_READ) as filecontent:
                for string in args:
                    if filecontent.find(string, self.offset(), self.offset() + len(self)) < 0:
                        return False
        return True

    @property
    def content(self) -> bytes:
        with self.save_pos():
            self.seek(0)
            return self.read(len(self))

    def tempfile(self, prefix=None, suffix=None):
        class FSTempfile:
            def __init__(self, file_stream: FileStream):
                self._temp = None
                self._fs: FileStream = file_stream

            def __enter__(self):
                self._temp = tf.NamedTemporaryFile(prefix=prefix, suffix=suffix, delete=False)
                self._temp.write(self._fs.content)
                self._temp.flush()
                self._temp.close()
                return self._temp.name

            def __exit__(self, type, value, traceback):
                if self._temp is not None:
                    os.unlink(self._temp.name)
                    self._temp = None
        return FSTempfile(self)

    def __getitem__(self, index):
        if isinstance(index, int):
            self.seek(index)
            return self.read(1)
        elif not isinstance(index, slice):
            raise ValueError(f"unexpected argument {index}")
        if index.step is not None and index.step != 1:
            raise ValueError(f"Invalid slice step: {index}")
        length=None
        if index.stop is not None:
            if index.stop < 0:
                length = len(self) + index.stop - index.start
            else:
                length = len(self) - (index.stop - index.start)
        return FileStream(self, start=index.start, length=length, close_on_exit=False)

    def __enter__(self) -> "FileStream":
        self._entries += 1
        return self

    def __exit__(self, type, value, traceback):
        self._entries -= 1
        assert self._entries >= 0
        if self._entries == 0 and self.close_on_exit:
            self.close_on_exit = False
            self._stream.close()



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class MozillaMar(KaitaiStruct):
    """Mozilla ARchive file is Mozilla's own archive format to distribute software updates.
    Test files can be found on Mozilla's FTP site, for example:
    
    <http://ftp.mozilla.org/pub/firefox/nightly/partials/>
    
    .. seealso::
       Source - https://wiki.mozilla.org/Software_Update:MAR
    """

    class SignatureAlgorithms(Enum):
        rsa_pkcs1_sha1 = 1
        rsa_pkcs1_sha384 = 2

    class BlockIdentifiers(Enum):
        product_information = 1
    SEQ_FIELDS = ["magic", "ofs_index", "file_size", "len_signatures", "signatures", "len_additional_sections", "additional_sections"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['magic']['start'] = self._io.pos()
        self.magic = self._io.read_bytes(4)
        self._debug['magic']['end'] = self._io.pos()
        if not self.magic == b"\x4D\x41\x52\x31":
            raise kaitaistruct.ValidationNotEqualError(b"\x4D\x41\x52\x31", self.magic, self._io, u"/seq/0")
        self._debug['ofs_index']['start'] = self._io.pos()
        self.ofs_index = self._io.read_u4be()
        self._debug['ofs_index']['end'] = self._io.pos()
        self._debug['file_size']['start'] = self._io.pos()
        self.file_size = self._io.read_u8be()
        self._debug['file_size']['end'] = self._io.pos()
        self._debug['len_signatures']['start'] = self._io.pos()
        self.len_signatures = self._io.read_u4be()
        self._debug['len_signatures']['end'] = self._io.pos()
        self._debug['signatures']['start'] = self._io.pos()
        self.signatures = [None] * (self.len_signatures)
        for i in range(self.len_signatures):
            if not 'arr' in self._debug['signatures']:
                self._debug['signatures']['arr'] = []
            self._debug['signatures']['arr'].append({'start': self._io.pos()})
            _t_signatures = MozillaMar.Signature(self._io, self, self._root)
            _t_signatures._read()
            self.signatures[i] = _t_signatures
            self._debug['signatures']['arr'][i]['end'] = self._io.pos()

        self._debug['signatures']['end'] = self._io.pos()
        self._debug['len_additional_sections']['start'] = self._io.pos()
        self.len_additional_sections = self._io.read_u4be()
        self._debug['len_additional_sections']['end'] = self._io.pos()
        self._debug['additional_sections']['start'] = self._io.pos()
        self.additional_sections = [None] * (self.len_additional_sections)
        for i in range(self.len_additional_sections):
            if not 'arr' in self._debug['additional_sections']:
                self._debug['additional_sections']['arr'] = []
            self._debug['additional_sections']['arr'].append({'start': self._io.pos()})
            _t_additional_sections = MozillaMar.AdditionalSection(self._io, self, self._root)
            _t_additional_sections._read()
            self.additional_sections[i] = _t_additional_sections
            self._debug['additional_sections']['arr'][i]['end'] = self._io.pos()

        self._debug['additional_sections']['end'] = self._io.pos()

    class MarIndex(KaitaiStruct):
        SEQ_FIELDS = ["len_index", "index_entries"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_index']['start'] = self._io.pos()
            self.len_index = self._io.read_u4be()
            self._debug['len_index']['end'] = self._io.pos()
            self._debug['index_entries']['start'] = self._io.pos()
            self._raw_index_entries = self._io.read_bytes(self.len_index)
            _io__raw_index_entries = KaitaiStream(BytesIO(self._raw_index_entries))
            self.index_entries = MozillaMar.IndexEntries(_io__raw_index_entries, self, self._root)
            self.index_entries._read()
            self._debug['index_entries']['end'] = self._io.pos()


    class IndexEntries(KaitaiStruct):
        SEQ_FIELDS = ["index_entry"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['index_entry']['start'] = self._io.pos()
            self.index_entry = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['index_entry']:
                    self._debug['index_entry']['arr'] = []
                self._debug['index_entry']['arr'].append({'start': self._io.pos()})
                _t_index_entry = MozillaMar.IndexEntry(self._io, self, self._root)
                _t_index_entry._read()
                self.index_entry.append(_t_index_entry)
                self._debug['index_entry']['arr'][len(self.index_entry) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['index_entry']['end'] = self._io.pos()


    class Signature(KaitaiStruct):
        SEQ_FIELDS = ["algorithm", "len_signature", "signature"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['algorithm']['start'] = self._io.pos()
            self.algorithm = KaitaiStream.resolve_enum(MozillaMar.SignatureAlgorithms, self._io.read_u4be())
            self._debug['algorithm']['end'] = self._io.pos()
            self._debug['len_signature']['start'] = self._io.pos()
            self.len_signature = self._io.read_u4be()
            self._debug['len_signature']['end'] = self._io.pos()
            self._debug['signature']['start'] = self._io.pos()
            self.signature = self._io.read_bytes(self.len_signature)
            self._debug['signature']['end'] = self._io.pos()


    class ProductInformationBlock(KaitaiStruct):
        SEQ_FIELDS = ["mar_channel_name", "product_version"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['mar_channel_name']['start'] = self._io.pos()
            self.mar_channel_name = (KaitaiStream.bytes_terminate(self._io.read_bytes(64), 0, False)).decode(u"UTF-8")
            self._debug['mar_channel_name']['end'] = self._io.pos()
            self._debug['product_version']['start'] = self._io.pos()
            self.product_version = (KaitaiStream.bytes_terminate(self._io.read_bytes(32), 0, False)).decode(u"UTF-8")
            self._debug['product_version']['end'] = self._io.pos()


    class IndexEntry(KaitaiStruct):
        SEQ_FIELDS = ["ofs_content", "len_content", "flags", "file_name"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['ofs_content']['start'] = self._io.pos()
            self.ofs_content = self._io.read_u4be()
            self._debug['ofs_content']['end'] = self._io.pos()
            self._debug['len_content']['start'] = self._io.pos()
            self.len_content = self._io.read_u4be()
            self._debug['len_content']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = self._io.read_u4be()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['file_name']['start'] = self._io.pos()
            self.file_name = (self._io.read_bytes_term(0, False, True, True)).decode(u"UTF-8")
            self._debug['file_name']['end'] = self._io.pos()

        @property
        def body(self):
            if hasattr(self, '_m_body'):
                return self._m_body if hasattr(self, '_m_body') else None

            io = self._root._io
            _pos = io.pos()
            io.seek(self.ofs_content)
            self._debug['_m_body']['start'] = io.pos()
            self._m_body = io.read_bytes(self.len_content)
            self._debug['_m_body']['end'] = io.pos()
            io.seek(_pos)
            return self._m_body if hasattr(self, '_m_body') else None


    class AdditionalSection(KaitaiStruct):
        SEQ_FIELDS = ["len_block", "block_identifier", "bytes"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_block']['start'] = self._io.pos()
            self.len_block = self._io.read_u4be()
            self._debug['len_block']['end'] = self._io.pos()
            self._debug['block_identifier']['start'] = self._io.pos()
            self.block_identifier = KaitaiStream.resolve_enum(MozillaMar.BlockIdentifiers, self._io.read_u4be())
            self._debug['block_identifier']['end'] = self._io.pos()
            self._debug['bytes']['start'] = self._io.pos()
            _on = self.block_identifier
            if _on == MozillaMar.BlockIdentifiers.product_information:
                self._raw_bytes = self._io.read_bytes(((self.len_block - 4) - 4))
                _io__raw_bytes = KaitaiStream(BytesIO(self._raw_bytes))
                self.bytes = MozillaMar.ProductInformationBlock(_io__raw_bytes, self, self._root)
                self.bytes._read()
            else:
                self.bytes = self._io.read_bytes(((self.len_block - 4) - 4))
            self._debug['bytes']['end'] = self._io.pos()


    @property
    def index(self):
        if hasattr(self, '_m_index'):
            return self._m_index if hasattr(self, '_m_index') else None

        _pos = self._io.pos()
        self._io.seek(self.ofs_index)
        self._debug['_m_index']['start'] = self._io.pos()
        self._m_index = MozillaMar.MarIndex(self._io, self, self._root)
        self._m_index._read()
        self._debug['_m_index']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_index if hasattr(self, '_m_index') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections
from enum import Enum


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class AndroidSuper(KaitaiStruct):
    """The metadata stored by Android at the beginning of a "super" partition, which
    is what it calls a disk partition that holds one or more Dynamic Partitions.
    Dynamic Partitions do more or less the same thing that LVM does on Linux,
    allowing Android to map ranges of non-contiguous extents to a single logical
    device. This metadata holds that mapping.
    
    .. seealso::
       Source - https://source.android.com/devices/tech/ota/dynamic_partitions
    
    
    .. seealso::
       Source - https://android.googlesource.com/platform/system/core/+/refs/tags/android-11.0.0_r8/fs_mgr/liblp/include/liblp/metadata_format.h
    """
    SEQ_FIELDS = []
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        pass

    class Root(KaitaiStruct):
        SEQ_FIELDS = ["primary_geometry", "backup_geometry", "primary_metadata", "backup_metadata"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['primary_geometry']['start'] = self._io.pos()
            self._raw_primary_geometry = self._io.read_bytes(4096)
            _io__raw_primary_geometry = KaitaiStream(BytesIO(self._raw_primary_geometry))
            self.primary_geometry = AndroidSuper.Geometry(_io__raw_primary_geometry, self, self._root)
            self.primary_geometry._read()
            self._debug['primary_geometry']['end'] = self._io.pos()
            self._debug['backup_geometry']['start'] = self._io.pos()
            self._raw_backup_geometry = self._io.read_bytes(4096)
            _io__raw_backup_geometry = KaitaiStream(BytesIO(self._raw_backup_geometry))
            self.backup_geometry = AndroidSuper.Geometry(_io__raw_backup_geometry, self, self._root)
            self.backup_geometry._read()
            self._debug['backup_geometry']['end'] = self._io.pos()
            self._debug['primary_metadata']['start'] = self._io.pos()
            self._raw_primary_metadata = [None] * (self.primary_geometry.metadata_slot_count)
            self.primary_metadata = [None] * (self.primary_geometry.metadata_slot_count)
            for i in range(self.primary_geometry.metadata_slot_count):
                if not 'arr' in self._debug['primary_metadata']:
                    self._debug['primary_metadata']['arr'] = []
                self._debug['primary_metadata']['arr'].append({'start': self._io.pos()})
                self._raw_primary_metadata[i] = self._io.read_bytes(self.primary_geometry.metadata_max_size)
                _io__raw_primary_metadata = KaitaiStream(BytesIO(self._raw_primary_metadata[i]))
                _t_primary_metadata = AndroidSuper.Metadata(_io__raw_primary_metadata, self, self._root)
                _t_primary_metadata._read()
                self.primary_metadata[i] = _t_primary_metadata
                self._debug['primary_metadata']['arr'][i]['end'] = self._io.pos()

            self._debug['primary_metadata']['end'] = self._io.pos()
            self._debug['backup_metadata']['start'] = self._io.pos()
            self._raw_backup_metadata = [None] * (self.primary_geometry.metadata_slot_count)
            self.backup_metadata = [None] * (self.primary_geometry.metadata_slot_count)
            for i in range(self.primary_geometry.metadata_slot_count):
                if not 'arr' in self._debug['backup_metadata']:
                    self._debug['backup_metadata']['arr'] = []
                self._debug['backup_metadata']['arr'].append({'start': self._io.pos()})
                self._raw_backup_metadata[i] = self._io.read_bytes(self.primary_geometry.metadata_max_size)
                _io__raw_backup_metadata = KaitaiStream(BytesIO(self._raw_backup_metadata[i]))
                _t_backup_metadata = AndroidSuper.Metadata(_io__raw_backup_metadata, self, self._root)
                _t_backup_metadata._read()
                self.backup_metadata[i] = _t_backup_metadata
                self._debug['backup_metadata']['arr'][i]['end'] = self._io.pos()

            self._debug['backup_metadata']['end'] = self._io.pos()


    class Geometry(KaitaiStruct):
        SEQ_FIELDS = ["magic", "struct_size", "checksum", "metadata_max_size", "metadata_slot_count", "logical_block_size"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x67\x44\x6C\x61":
                raise kaitaistruct.ValidationNotEqualError(b"\x67\x44\x6C\x61", self.magic, self._io, u"/types/geometry/seq/0")
            self._debug['struct_size']['start'] = self._io.pos()
            self.struct_size = self._io.read_u4le()
            self._debug['struct_size']['end'] = self._io.pos()
            self._debug['checksum']['start'] = self._io.pos()
            self.checksum = self._io.read_bytes(32)
            self._debug['checksum']['end'] = self._io.pos()
            self._debug['metadata_max_size']['start'] = self._io.pos()
            self.metadata_max_size = self._io.read_u4le()
            self._debug['metadata_max_size']['end'] = self._io.pos()
            self._debug['metadata_slot_count']['start'] = self._io.pos()
            self.metadata_slot_count = self._io.read_u4le()
            self._debug['metadata_slot_count']['end'] = self._io.pos()
            self._debug['logical_block_size']['start'] = self._io.pos()
            self.logical_block_size = self._io.read_u4le()
            self._debug['logical_block_size']['end'] = self._io.pos()


    class Metadata(KaitaiStruct):

        class TableKind(Enum):
            partitions = 0
            extents = 1
            groups = 2
            block_devices = 3
        SEQ_FIELDS = ["magic", "major_version", "minor_version", "header_size", "header_checksum", "tables_size", "tables_checksum", "partitions", "extents", "groups", "block_devices"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x30\x50\x4C\x41":
                raise kaitaistruct.ValidationNotEqualError(b"\x30\x50\x4C\x41", self.magic, self._io, u"/types/metadata/seq/0")
            self._debug['major_version']['start'] = self._io.pos()
            self.major_version = self._io.read_u2le()
            self._debug['major_version']['end'] = self._io.pos()
            self._debug['minor_version']['start'] = self._io.pos()
            self.minor_version = self._io.read_u2le()
            self._debug['minor_version']['end'] = self._io.pos()
            self._debug['header_size']['start'] = self._io.pos()
            self.header_size = self._io.read_u4le()
            self._debug['header_size']['end'] = self._io.pos()
            self._debug['header_checksum']['start'] = self._io.pos()
            self.header_checksum = self._io.read_bytes(32)
            self._debug['header_checksum']['end'] = self._io.pos()
            self._debug['tables_size']['start'] = self._io.pos()
            self.tables_size = self._io.read_u4le()
            self._debug['tables_size']['end'] = self._io.pos()
            self._debug['tables_checksum']['start'] = self._io.pos()
            self.tables_checksum = self._io.read_bytes(32)
            self._debug['tables_checksum']['end'] = self._io.pos()
            self._debug['partitions']['start'] = self._io.pos()
            self.partitions = AndroidSuper.Metadata.TableDescriptor(AndroidSuper.Metadata.TableKind.partitions, self._io, self, self._root)
            self.partitions._read()
            self._debug['partitions']['end'] = self._io.pos()
            self._debug['extents']['start'] = self._io.pos()
            self.extents = AndroidSuper.Metadata.TableDescriptor(AndroidSuper.Metadata.TableKind.extents, self._io, self, self._root)
            self.extents._read()
            self._debug['extents']['end'] = self._io.pos()
            self._debug['groups']['start'] = self._io.pos()
            self.groups = AndroidSuper.Metadata.TableDescriptor(AndroidSuper.Metadata.TableKind.groups, self._io, self, self._root)
            self.groups._read()
            self._debug['groups']['end'] = self._io.pos()
            self._debug['block_devices']['start'] = self._io.pos()
            self.block_devices = AndroidSuper.Metadata.TableDescriptor(AndroidSuper.Metadata.TableKind.block_devices, self._io, self, self._root)
            self.block_devices._read()
            self._debug['block_devices']['end'] = self._io.pos()

        class BlockDevice(KaitaiStruct):
            SEQ_FIELDS = ["first_logical_sector", "alignment", "alignment_offset", "size", "partition_name", "flag_slot_suffixed", "flags_reserved"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['first_logical_sector']['start'] = self._io.pos()
                self.first_logical_sector = self._io.read_u8le()
                self._debug['first_logical_sector']['end'] = self._io.pos()
                self._debug['alignment']['start'] = self._io.pos()
                self.alignment = self._io.read_u4le()
                self._debug['alignment']['end'] = self._io.pos()
                self._debug['alignment_offset']['start'] = self._io.pos()
                self.alignment_offset = self._io.read_u4le()
                self._debug['alignment_offset']['end'] = self._io.pos()
                self._debug['size']['start'] = self._io.pos()
                self.size = self._io.read_u8le()
                self._debug['size']['end'] = self._io.pos()
                self._debug['partition_name']['start'] = self._io.pos()
                self.partition_name = (KaitaiStream.bytes_terminate(self._io.read_bytes(36), 0, False)).decode(u"UTF-8")
                self._debug['partition_name']['end'] = self._io.pos()
                self._debug['flag_slot_suffixed']['start'] = self._io.pos()
                self.flag_slot_suffixed = self._io.read_bits_int_le(1) != 0
                self._debug['flag_slot_suffixed']['end'] = self._io.pos()
                self._debug['flags_reserved']['start'] = self._io.pos()
                self.flags_reserved = self._io.read_bits_int_le(31)
                self._debug['flags_reserved']['end'] = self._io.pos()


        class Extent(KaitaiStruct):

            class TargetType(Enum):
                linear = 0
                zero = 1
            SEQ_FIELDS = ["num_sectors", "target_type", "target_data", "target_source"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['num_sectors']['start'] = self._io.pos()
                self.num_sectors = self._io.read_u8le()
                self._debug['num_sectors']['end'] = self._io.pos()
                self._debug['target_type']['start'] = self._io.pos()
                self.target_type = KaitaiStream.resolve_enum(AndroidSuper.Metadata.Extent.TargetType, self._io.read_u4le())
                self._debug['target_type']['end'] = self._io.pos()
                self._debug['target_data']['start'] = self._io.pos()
                self.target_data = self._io.read_u8le()
                self._debug['target_data']['end'] = self._io.pos()
                self._debug['target_source']['start'] = self._io.pos()
                self.target_source = self._io.read_u4le()
                self._debug['target_source']['end'] = self._io.pos()


        class TableDescriptor(KaitaiStruct):
            SEQ_FIELDS = ["offset", "num_entries", "entry_size"]
            def __init__(self, kind, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self.kind = kind
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['offset']['start'] = self._io.pos()
                self.offset = self._io.read_u4le()
                self._debug['offset']['end'] = self._io.pos()
                self._debug['num_entries']['start'] = self._io.pos()
                self.num_entries = self._io.read_u4le()
                self._debug['num_entries']['end'] = self._io.pos()
                self._debug['entry_size']['start'] = self._io.pos()
                self.entry_size = self._io.read_u4le()
                self._debug['entry_size']['end'] = self._io.pos()

            @property
            def table(self):
                if hasattr(self, '_m_table'):
                    return self._m_table if hasattr(self, '_m_table') else None

                _pos = self._io.pos()
                self._io.seek((self._parent.header_size + self.offset))
                self._debug['_m_table']['start'] = self._io.pos()
                self._raw__m_table = [None] * (self.num_entries)
                self._m_table = [None] * (self.num_entries)
                for i in range(self.num_entries):
                    if not 'arr' in self._debug['_m_table']:
                        self._debug['_m_table']['arr'] = []
                    self._debug['_m_table']['arr'].append({'start': self._io.pos()})
                    _on = self.kind
                    if _on == AndroidSuper.Metadata.TableKind.partitions:
                        if not 'arr' in self._debug['_m_table']:
                            self._debug['_m_table']['arr'] = []
                        self._debug['_m_table']['arr'].append({'start': self._io.pos()})
                        self._raw__m_table[i] = self._io.read_bytes(self.entry_size)
                        _io__raw__m_table = KaitaiStream(BytesIO(self._raw__m_table[i]))
                        _t__m_table = AndroidSuper.Metadata.Partition(_io__raw__m_table, self, self._root)
                        _t__m_table._read()
                        self._m_table[i] = _t__m_table
                        self._debug['_m_table']['arr'][i]['end'] = self._io.pos()
                    elif _on == AndroidSuper.Metadata.TableKind.extents:
                        if not 'arr' in self._debug['_m_table']:
                            self._debug['_m_table']['arr'] = []
                        self._debug['_m_table']['arr'].append({'start': self._io.pos()})
                        self._raw__m_table[i] = self._io.read_bytes(self.entry_size)
                        _io__raw__m_table = KaitaiStream(BytesIO(self._raw__m_table[i]))
                        _t__m_table = AndroidSuper.Metadata.Extent(_io__raw__m_table, self, self._root)
                        _t__m_table._read()
                        self._m_table[i] = _t__m_table
                        self._debug['_m_table']['arr'][i]['end'] = self._io.pos()
                    elif _on == AndroidSuper.Metadata.TableKind.groups:
                        if not 'arr' in self._debug['_m_table']:
                            self._debug['_m_table']['arr'] = []
                        self._debug['_m_table']['arr'].append({'start': self._io.pos()})
                        self._raw__m_table[i] = self._io.read_bytes(self.entry_size)
                        _io__raw__m_table = KaitaiStream(BytesIO(self._raw__m_table[i]))
                        _t__m_table = AndroidSuper.Metadata.Group(_io__raw__m_table, self, self._root)
                        _t__m_table._read()
                        self._m_table[i] = _t__m_table
                        self._debug['_m_table']['arr'][i]['end'] = self._io.pos()
                    elif _on == AndroidSuper.Metadata.TableKind.block_devices:
                        if not 'arr' in self._debug['_m_table']:
                            self._debug['_m_table']['arr'] = []
                        self._debug['_m_table']['arr'].append({'start': self._io.pos()})
                        self._raw__m_table[i] = self._io.read_bytes(self.entry_size)
                        _io__raw__m_table = KaitaiStream(BytesIO(self._raw__m_table[i]))
                        _t__m_table = AndroidSuper.Metadata.BlockDevice(_io__raw__m_table, self, self._root)
                        _t__m_table._read()
                        self._m_table[i] = _t__m_table
                        self._debug['_m_table']['arr'][i]['end'] = self._io.pos()
                    else:
                        if not 'arr' in self._debug['_m_table']:
                            self._debug['_m_table']['arr'] = []
                        self._debug['_m_table']['arr'].append({'start': self._io.pos()})
                        self._m_table[i] = self._io.read_bytes(self.entry_size)
                        self._debug['_m_table']['arr'][i]['end'] = self._io.pos()
                    self._debug['_m_table']['arr'][i]['end'] = self._io.pos()

                self._debug['_m_table']['end'] = self._io.pos()
                self._io.seek(_pos)
                return self._m_table if hasattr(self, '_m_table') else None


        class Partition(KaitaiStruct):
            SEQ_FIELDS = ["name", "attr_readonly", "attr_slot_suffixed", "attr_updated", "attr_disabled", "attrs_reserved", "first_extent_index", "num_extents", "group_index"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['name']['start'] = self._io.pos()
                self.name = (KaitaiStream.bytes_terminate(self._io.read_bytes(36), 0, False)).decode(u"UTF-8")
                self._debug['name']['end'] = self._io.pos()
                self._debug['attr_readonly']['start'] = self._io.pos()
                self.attr_readonly = self._io.read_bits_int_le(1) != 0
                self._debug['attr_readonly']['end'] = self._io.pos()
                self._debug['attr_slot_suffixed']['start'] = self._io.pos()
                self.attr_slot_suffixed = self._io.read_bits_int_le(1) != 0
                self._debug['attr_slot_suffixed']['end'] = self._io.pos()
                self._debug['attr_updated']['start'] = self._io.pos()
                self.attr_updated = self._io.read_bits_int_le(1) != 0
                self._debug['attr_updated']['end'] = self._io.pos()
                self._debug['attr_disabled']['start'] = self._io.pos()
                self.attr_disabled = self._io.read_bits_int_le(1) != 0
                self._debug['attr_disabled']['end'] = self._io.pos()
                self._debug['attrs_reserved']['start'] = self._io.pos()
                self.attrs_reserved = self._io.read_bits_int_le(28)
                self._debug['attrs_reserved']['end'] = self._io.pos()
                self._io.align_to_byte()
                self._debug['first_extent_index']['start'] = self._io.pos()
                self.first_extent_index = self._io.read_u4le()
                self._debug['first_extent_index']['end'] = self._io.pos()
                self._debug['num_extents']['start'] = self._io.pos()
                self.num_extents = self._io.read_u4le()
                self._debug['num_extents']['end'] = self._io.pos()
                self._debug['group_index']['start'] = self._io.pos()
                self.group_index = self._io.read_u4le()
                self._debug['group_index']['end'] = self._io.pos()


        class Group(KaitaiStruct):
            SEQ_FIELDS = ["name", "flag_slot_suffixed", "flags_reserved", "maximum_size"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['name']['start'] = self._io.pos()
                self.name = (KaitaiStream.bytes_terminate(self._io.read_bytes(36), 0, False)).decode(u"UTF-8")
                self._debug['name']['end'] = self._io.pos()
                self._debug['flag_slot_suffixed']['start'] = self._io.pos()
                self.flag_slot_suffixed = self._io.read_bits_int_le(1) != 0
                self._debug['flag_slot_suffixed']['end'] = self._io.pos()
                self._debug['flags_reserved']['start'] = self._io.pos()
                self.flags_reserved = self._io.read_bits_int_le(31)
                self._debug['flags_reserved']['end'] = self._io.pos()
                self._io.align_to_byte()
                self._debug['maximum_size']['start'] = self._io.pos()
                self.maximum_size = self._io.read_u8le()
                self._debug['maximum_size']['end'] = self._io.pos()



    @property
    def root(self):
        if hasattr(self, '_m_root'):
            return self._m_root if hasattr(self, '_m_root') else None

        _pos = self._io.pos()
        self._io.seek(4096)
        self._debug['_m_root']['start'] = self._io.pos()
        self._m_root = AndroidSuper.Root(self._io, self, self._root)
        self._m_root._read()
        self._debug['_m_root']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_root if hasattr(self, '_m_root') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections
from enum import Enum


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Bson(KaitaiStruct):
    """BSON, short for Binary JSON, is a binary-encoded serialization of JSON-like documents. Like JSON, BSON supports the embedding of documents and arrays within other documents and arrays. BSON also contains extensions that allow representation of data types that are not part of the JSON spec. For example, BSON has a Date type and a BinData type. BSON can be compared to binary interchange formats, like Protocol Buffers. BSON is more "schemaless" than Protocol Buffers, which can give it an advantage in flexibility but also a slight disadvantage in space efficiency (BSON has overhead for field names within the serialized data). BSON was designed to have the following three characteristics:
      * Lightweight. Keeping spatial overhead to a minimum is important for any data representation format, especially when used over the network.
      * Traversable. BSON is designed to be traversed easily. This is a vital property in its role as the primary data representation for MongoDB.
      * Efficient. Encoding data to BSON and decoding from BSON can be performed very quickly in most languages due to the use of C data types.
    """
    SEQ_FIELDS = ["len", "fields", "terminator"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['len']['start'] = self._io.pos()
        self.len = self._io.read_s4le()
        self._debug['len']['end'] = self._io.pos()
        self._debug['fields']['start'] = self._io.pos()
        self._raw_fields = self._io.read_bytes((self.len - 5))
        _io__raw_fields = KaitaiStream(BytesIO(self._raw_fields))
        self.fields = Bson.ElementsList(_io__raw_fields, self, self._root)
        self.fields._read()
        self._debug['fields']['end'] = self._io.pos()
        self._debug['terminator']['start'] = self._io.pos()
        self.terminator = self._io.read_bytes(1)
        self._debug['terminator']['end'] = self._io.pos()
        if not self.terminator == b"\x00":
            raise kaitaistruct.ValidationNotEqualError(b"\x00", self.terminator, self._io, u"/seq/2")

    class Timestamp(KaitaiStruct):
        """Special internal type used by MongoDB replication and sharding. First 4 bytes are an increment, second 4 are a timestamp."""
        SEQ_FIELDS = ["increment", "timestamp"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['increment']['start'] = self._io.pos()
            self.increment = self._io.read_u4le()
            self._debug['increment']['end'] = self._io.pos()
            self._debug['timestamp']['start'] = self._io.pos()
            self.timestamp = self._io.read_u4le()
            self._debug['timestamp']['end'] = self._io.pos()


    class BinData(KaitaiStruct):
        """The BSON "binary" or "BinData" datatype is used to represent arrays of bytes. It is somewhat analogous to the Java notion of a ByteArray. BSON binary values have a subtype. This is used to indicate what kind of data is in the byte array. Subtypes from zero to 127 are predefined or reserved. Subtypes from 128-255 are user-defined."""

        class Subtype(Enum):
            generic = 0
            function = 1
            byte_array_deprecated = 2
            uuid_deprecated = 3
            uuid = 4
            md5 = 5
            custom = 128
        SEQ_FIELDS = ["len", "subtype", "content"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len']['start'] = self._io.pos()
            self.len = self._io.read_s4le()
            self._debug['len']['end'] = self._io.pos()
            self._debug['subtype']['start'] = self._io.pos()
            self.subtype = KaitaiStream.resolve_enum(Bson.BinData.Subtype, self._io.read_u1())
            self._debug['subtype']['end'] = self._io.pos()
            self._debug['content']['start'] = self._io.pos()
            _on = self.subtype
            if _on == Bson.BinData.Subtype.byte_array_deprecated:
                self._raw_content = self._io.read_bytes(self.len)
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = Bson.BinData.ByteArrayDeprecated(_io__raw_content, self, self._root)
                self.content._read()
            else:
                self.content = self._io.read_bytes(self.len)
            self._debug['content']['end'] = self._io.pos()

        class ByteArrayDeprecated(KaitaiStruct):
            """The BSON "binary" or "BinData" datatype is used to represent arrays of bytes. It is somewhat analogous to the Java notion of a ByteArray. BSON binary values have a subtype. This is used to indicate what kind of data is in the byte array. Subtypes from zero to 127 are predefined or reserved. Subtypes from 128-255 are user-defined."""
            SEQ_FIELDS = ["len", "content"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['len']['start'] = self._io.pos()
                self.len = self._io.read_s4le()
                self._debug['len']['end'] = self._io.pos()
                self._debug['content']['start'] = self._io.pos()
                self.content = self._io.read_bytes(self.len)
                self._debug['content']['end'] = self._io.pos()



    class ElementsList(KaitaiStruct):
        SEQ_FIELDS = ["elements"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['elements']['start'] = self._io.pos()
            self.elements = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['elements']:
                    self._debug['elements']['arr'] = []
                self._debug['elements']['arr'].append({'start': self._io.pos()})
                _t_elements = Bson.Element(self._io, self, self._root)
                _t_elements._read()
                self.elements.append(_t_elements)
                self._debug['elements']['arr'][len(self.elements) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['elements']['end'] = self._io.pos()


    class Cstring(KaitaiStruct):
        SEQ_FIELDS = ["str"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['str']['start'] = self._io.pos()
            self.str = (self._io.read_bytes_term(0, False, True, True)).decode(u"UTF-8")
            self._debug['str']['end'] = self._io.pos()


    class String(KaitaiStruct):
        SEQ_FIELDS = ["len", "str", "terminator"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len']['start'] = self._io.pos()
            self.len = self._io.read_s4le()
            self._debug['len']['end'] = self._io.pos()
            self._debug['str']['start'] = self._io.pos()
            self.str = (self._io.read_bytes((self.len - 1))).decode(u"UTF-8")
            self._debug['str']['end'] = self._io.pos()
            self._debug['terminator']['start'] = self._io.pos()
            self.terminator = self._io.read_bytes(1)
            self._debug['terminator']['end'] = self._io.pos()
            if not self.terminator == b"\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x00", self.terminator, self._io, u"/types/string/seq/2")


    class Element(KaitaiStruct):

        class BsonType(Enum):
            min_key = -1
            end_of_object = 0
            number_double = 1
            string = 2
            document = 3
            array = 4
            bin_data = 5
            undefined = 6
            object_id = 7
            boolean = 8
            utc_datetime = 9
            jst_null = 10
            reg_ex = 11
            db_pointer = 12
            javascript = 13
            symbol = 14
            code_with_scope = 15
            number_int = 16
            timestamp = 17
            number_long = 18
            number_decimal = 19
            max_key = 127
        SEQ_FIELDS = ["type_byte", "name", "content"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['type_byte']['start'] = self._io.pos()
            self.type_byte = KaitaiStream.resolve_enum(Bson.Element.BsonType, self._io.read_u1())
            self._debug['type_byte']['end'] = self._io.pos()
            self._debug['name']['start'] = self._io.pos()
            self.name = Bson.Cstring(self._io, self, self._root)
            self.name._read()
            self._debug['name']['end'] = self._io.pos()
            self._debug['content']['start'] = self._io.pos()
            _on = self.type_byte
            if _on == Bson.Element.BsonType.code_with_scope:
                self.content = Bson.CodeWithScope(self._io, self, self._root)
                self.content._read()
            elif _on == Bson.Element.BsonType.reg_ex:
                self.content = Bson.RegEx(self._io, self, self._root)
                self.content._read()
            elif _on == Bson.Element.BsonType.number_double:
                self.content = self._io.read_f8le()
            elif _on == Bson.Element.BsonType.symbol:
                self.content = Bson.String(self._io, self, self._root)
                self.content._read()
            elif _on == Bson.Element.BsonType.timestamp:
                self.content = Bson.Timestamp(self._io, self, self._root)
                self.content._read()
            elif _on == Bson.Element.BsonType.number_int:
                self.content = self._io.read_s4le()
            elif _on == Bson.Element.BsonType.document:
                self.content = Bson(self._io)
                self.content._read()
            elif _on == Bson.Element.BsonType.object_id:
                self.content = Bson.ObjectId(self._io, self, self._root)
                self.content._read()
            elif _on == Bson.Element.BsonType.javascript:
                self.content = Bson.String(self._io, self, self._root)
                self.content._read()
            elif _on == Bson.Element.BsonType.utc_datetime:
                self.content = self._io.read_s8le()
            elif _on == Bson.Element.BsonType.boolean:
                self.content = self._io.read_u1()
            elif _on == Bson.Element.BsonType.number_long:
                self.content = self._io.read_s8le()
            elif _on == Bson.Element.BsonType.bin_data:
                self.content = Bson.BinData(self._io, self, self._root)
                self.content._read()
            elif _on == Bson.Element.BsonType.string:
                self.content = Bson.String(self._io, self, self._root)
                self.content._read()
            elif _on == Bson.Element.BsonType.db_pointer:
                self.content = Bson.DbPointer(self._io, self, self._root)
                self.content._read()
            elif _on == Bson.Element.BsonType.array:
                self.content = Bson(self._io)
                self.content._read()
            elif _on == Bson.Element.BsonType.number_decimal:
                self.content = Bson.F16(self._io, self, self._root)
                self.content._read()
            self._debug['content']['end'] = self._io.pos()


    class DbPointer(KaitaiStruct):
        SEQ_FIELDS = ["namespace", "id"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['namespace']['start'] = self._io.pos()
            self.namespace = Bson.String(self._io, self, self._root)
            self.namespace._read()
            self._debug['namespace']['end'] = self._io.pos()
            self._debug['id']['start'] = self._io.pos()
            self.id = Bson.ObjectId(self._io, self, self._root)
            self.id._read()
            self._debug['id']['end'] = self._io.pos()


    class U3(KaitaiStruct):
        """Implements unsigned 24-bit (3 byte) integer.
        """
        SEQ_FIELDS = ["b1", "b2", "b3"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['b1']['start'] = self._io.pos()
            self.b1 = self._io.read_u1()
            self._debug['b1']['end'] = self._io.pos()
            self._debug['b2']['start'] = self._io.pos()
            self.b2 = self._io.read_u1()
            self._debug['b2']['end'] = self._io.pos()
            self._debug['b3']['start'] = self._io.pos()
            self.b3 = self._io.read_u1()
            self._debug['b3']['end'] = self._io.pos()

        @property
        def value(self):
            if hasattr(self, '_m_value'):
                return self._m_value if hasattr(self, '_m_value') else None

            self._m_value = ((self.b1 | (self.b2 << 8)) | (self.b3 << 16))
            return self._m_value if hasattr(self, '_m_value') else None


    class CodeWithScope(KaitaiStruct):
        SEQ_FIELDS = ["id", "source", "scope"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['id']['start'] = self._io.pos()
            self.id = self._io.read_s4le()
            self._debug['id']['end'] = self._io.pos()
            self._debug['source']['start'] = self._io.pos()
            self.source = Bson.String(self._io, self, self._root)
            self.source._read()
            self._debug['source']['end'] = self._io.pos()
            self._debug['scope']['start'] = self._io.pos()
            self.scope = Bson(self._io)
            self.scope._read()
            self._debug['scope']['end'] = self._io.pos()


    class F16(KaitaiStruct):
        """128-bit IEEE 754-2008 decimal floating point."""
        SEQ_FIELDS = ["str", "exponent", "significand_hi", "significand_lo"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['str']['start'] = self._io.pos()
            self.str = self._io.read_bits_int_be(1) != 0
            self._debug['str']['end'] = self._io.pos()
            self._debug['exponent']['start'] = self._io.pos()
            self.exponent = self._io.read_bits_int_be(15)
            self._debug['exponent']['end'] = self._io.pos()
            self._debug['significand_hi']['start'] = self._io.pos()
            self.significand_hi = self._io.read_bits_int_be(49)
            self._debug['significand_hi']['end'] = self._io.pos()
            self._io.align_to_byte()
            self._debug['significand_lo']['start'] = self._io.pos()
            self.significand_lo = self._io.read_u8le()
            self._debug['significand_lo']['end'] = self._io.pos()


    class ObjectId(KaitaiStruct):
        """https://docs.mongodb.com/manual/reference/method/ObjectId/."""
        SEQ_FIELDS = ["epoch_time", "machine_id", "process_id", "counter"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['epoch_time']['start'] = self._io.pos()
            self.epoch_time = self._io.read_u4le()
            self._debug['epoch_time']['end'] = self._io.pos()
            self._debug['machine_id']['start'] = self._io.pos()
            self.machine_id = Bson.U3(self._io, self, self._root)
            self.machine_id._read()
            self._debug['machine_id']['end'] = self._io.pos()
            self._debug['process_id']['start'] = self._io.pos()
            self.process_id = self._io.read_u2le()
            self._debug['process_id']['end'] = self._io.pos()
            self._debug['counter']['start'] = self._io.pos()
            self.counter = Bson.U3(self._io, self, self._root)
            self.counter._read()
            self._debug['counter']['end'] = self._io.pos()


    class RegEx(KaitaiStruct):
        SEQ_FIELDS = ["pattern", "options"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['pattern']['start'] = self._io.pos()
            self.pattern = Bson.Cstring(self._io, self, self._root)
            self.pattern._read()
            self._debug['pattern']['end'] = self._io.pos()
            self._debug['options']['start'] = self._io.pos()
            self.options = Bson.Cstring(self._io, self, self._root)
            self.options._read()
            self._debug['options']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class HeroesOfMightAndMagicBmp(KaitaiStruct):
    SEQ_FIELDS = ["magic", "width", "height", "data"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['magic']['start'] = self._io.pos()
        self.magic = self._io.read_u2le()
        self._debug['magic']['end'] = self._io.pos()
        self._debug['width']['start'] = self._io.pos()
        self.width = self._io.read_u2le()
        self._debug['width']['end'] = self._io.pos()
        self._debug['height']['start'] = self._io.pos()
        self.height = self._io.read_u2le()
        self._debug['height']['end'] = self._io.pos()
        self._debug['data']['start'] = self._io.pos()
        self.data = self._io.read_bytes((self.width * self.height))
        self._debug['data']['end'] = self._io.pos()





from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
import importlib.util
import inspect
from io import BufferedReader, BytesIO
import json
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Set, Type, Union

from .compiler import CompiledKSY
from ..fileutils import FileStream

from kaitaistruct import KaitaiStruct

PARSER_DIR: Path = Path(__file__).absolute().parent / "parsers"
MANIFEST_FILE: Path = PARSER_DIR / "manifest.json"

with open(MANIFEST_FILE, "r") as f:
    MANIFEST: Dict[str, Dict[str, Any]] = json.load(f)
COMPILED_INFO_BY_KSY: Dict[str, CompiledKSY] = {
    ksy_path: CompiledKSY(
        class_name=component["class_name"],
        python_path=PARSER_DIR / component["python_path"],
        dependencies=(
            CompiledKSY(class_name=dep["class_name"], python_path=PARSER_DIR / dep["python_path"])
            for dep in component["dependencies"]
        )
    )
    for ksy_path, component in MANIFEST.items()
}
del MANIFEST

_IMPORTED_SPECS: Set[Path] = set()
_PARSERS_BY_KSY: Dict[str, Type[KaitaiStruct]] = {}


def import_spec(compiled: CompiledKSY) -> Optional[Type[KaitaiStruct]]:
    if compiled.python_path in _IMPORTED_SPECS:
        return None
    _IMPORTED_SPECS.add(compiled.python_path)
    module_name = compiled.python_path.name
    assert module_name.lower().endswith(".py")
    module_name = module_name[:-3]
    module_name = f"{__name__}.parsers.{module_name}"
    spec = importlib.util.spec_from_file_location(module_name, compiled.python_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    for objname, obj in inspect.getmembers(module):
        if inspect.isclass(obj) and objname == compiled.class_name:
            return obj
    raise ImportError(f"Could not find parser class {compiled.class_name!r} in {compiled.python_path}")


@dataclass(unsafe_hash=True, order=True, frozen=True)
class Segment:
    start: int
    end: int

    def __contains__(self, item):
        return isinstance(item, Segment) and item.start >= self.start and item.end <= self.end

    def __len__(self):
        return self.end - self.start

    def __bool__(self):
        return len(self) > 0

    def __getitem__(self, index_or_slice: Union[int, slice]):
        if isinstance(index_or_slice, slice):
            if index_or_slice.step is not None and index_or_slice.step != 1:
                raise ValueError(f"{self.__class__.__name__}.__getitem__ only supports slices with step=1")
            if index_or_slice.start is None:
                new_start = self.start
            else:
                new_start = self.start + index_or_slice.start
            if index_or_slice.stop is None:
                new_end = self.end
            elif index_or_slice.stop < 0:
                new_end = self.end + index_or_slice.stop
            else:
                new_end = self.start + index_or_slice.stop
            if new_start > self.end:
                new_start = self.end
            if new_end < new_start or new_end > self.end:
                new_end = new_start
            return Segment(new_start, new_end)
        elif self.start + index_or_slice >= self.end or (index_or_slice < 0 and -index_or_slice > len(self)):
            raise IndexError(index_or_slice)
        elif index_or_slice < 0:
            return Segment(self.end + index_or_slice, self.end + index_or_slice + 1)
        else:
            return Segment(self.start + index_or_slice, self.start + index_or_slice + 1)


class ASTNode:
    """Represents an element in a parse."""

    def __init__(
            self,
            name: str,
            segment: Segment,
            offset: int,
            parent: Optional["CompoundNode"] = None
    ):
        self.name: str = name
        self.segment: Segment = segment
        self.offset: int = offset
        if parent is None:
            if not isinstance(self, RootNode):
                raise ValueError(f"Only a RootNode can have no parent, not {self!r}")
            self.root: RootNode = self
            self.level: int = 0
        else:
            self.root = parent.root
            self.level = parent.level + 1
        self.parent: Optional[CompoundNode] = parent

    def dfs(self) -> Iterator["ASTNode"]:
        yield self

    @property
    def start(self) -> int:
        return self.offset + self.segment.start

    @property
    def end(self) -> int:
        return self.offset + self.segment.end

    @property
    def raw_value(self):
        """Sequence of bytes of this segment."""
        return self.root.get_value(self.start, self.end)

    @property
    def size(self) -> int:
        return self.end - self.start

    @property
    def children(self) -> List["ASTNode"]:
        return []

    def __repr__(self):
        return f"{self.name}({self.__class__.__name__}) [{self.start}:{self.end}]"


class ValueNode(ASTNode):
    """A leaf in the parse tree."""

    TYPES = (int, float, str, bytes, Enum)

    def __init__(self, value: bytes, *args, **kwargs):
        self._value: bytes = value
        super().__init__(*args, **kwargs)

    @property
    def value(self) -> bytes:
        return self._value

    def __repr__(self):
        return f"{self.name}({self.__class__.__name__}<{self.value.__class__.__name__}>) [{self.start}:{self.end}]"


class CompoundNode(ASTNode, ABC):
    """A node that can have children"""

    def __init__(self, obj: KaitaiStruct, *args, **kwargs):
        self.obj: KaitaiStruct = obj
        super().__init__(*args, **kwargs)
        self._children: Optional[List[ASTNode]] = None

    @property
    def children(self) -> List[ASTNode]:
        if self._children is None:
            self._children = list(self.explore())
        return self._children

    @abstractmethod
    def explore(self) -> Iterator[ASTNode]:
        raise NotImplementedError()

    def dfs(self) -> Iterator[ASTNode]:
        stack = [self]
        while stack:
            top = stack.pop()
            yield top
            stack.extend(reversed(top.children))

    def make_child(
            self,
            obj: KaitaiStruct,
            name: str,
            segment: Segment,
            offset: int,
    ) -> ASTNode:
        if isinstance(obj, KaitaiStruct):
            node_class = StructNode
        elif isinstance(obj, ValueNode.TYPES):
            node_class = ValueNode
        elif isinstance(obj, list):
            node_class = ArrayNode
        else:
            raise TypeError(f"Unknown object type: {type(obj)}")

        return node_class(obj, name, segment, offset, self)


class StructNode(CompoundNode):
    """Represents node of the subtype."""

    def explore(self) -> Iterator[ASTNode]:
        for name in self.obj.SEQ_FIELDS:
            markers = self.obj._debug[name].copy()
            if "arr" in markers:
                del markers["arr"]
            if "start" not in markers or "end" not in markers:
                continue
            segment = Segment(**markers)
            offset = self.offset
            if isinstance(self.parent, StructNode):
                if self.obj._io != self.parent.obj._io:
                    offset = self.start
            if hasattr(self.obj, name):
                yield self.make_child(getattr(self.obj, name), name, segment, offset)


class ArrayNode(CompoundNode):
    """Represents node of array of subtype items."""

    def explore(self) -> Iterator[ASTNode]:
        for i, obj in enumerate(self.obj):
            markers = self.parent.obj._debug[self.name]["arr"][i]
            segment = Segment(**markers)
            name = f"{self.name}[{i}]"
            yield self.make_child(obj, name, segment, self.offset)


class RootNode(StructNode):
    def __init__(self, buffer: bytes, obj: KaitaiStruct):
        self.buffer: bytes = buffer
        super().__init__(obj, name=obj.__class__.__name__, segment=Segment(0, len(self.buffer)), offset=0)

    def get_value(self, start, end):
        return self.buffer[start:end]


class KaitaiInspector:
    def __init__(self, struct: KaitaiStruct):
        self.struct: KaitaiStruct = struct
        self._ast: Optional[RootNode] = None

    @property
    def ast(self) -> RootNode:
        if self._ast is None:
            _io = self.struct._io._io
            if isinstance(_io, FileStream):
                buffer = _io.content
            elif isinstance(_io, BytesIO):
                buffer = _io.getbuffer().tobytes()
            elif isinstance(_io, BufferedReader):
                with open(self.struct._io._io.name, 'rb') as f:
                    buffer = f.read()
            else:
                raise TypeError('Unsupported stream type')

            self._ast = RootNode(buffer, self.struct)

        return self._ast


class KaitaiParser:
    def __init__(self, struct_type: Type[KaitaiStruct]):
        self.struct_type: Type[KaitaiStruct] = struct_type

    @staticmethod
    def load(ksy_path: str) -> "KaitaiParser":
        """Returns a parser for the given KSY file and input file.

        The KSY file is specified as a relative path to the file within the Kaitai struct format library.
        Examples:

             "executable/dos_mz.ksy"
             "image/jpeg.ksy"
             "network/pcap.ksy"

        """
        if ksy_path not in _PARSERS_BY_KSY:
            if ksy_path not in COMPILED_INFO_BY_KSY:
                raise KeyError(ksy_path)
            info = COMPILED_INFO_BY_KSY[ksy_path]
            _PARSERS_BY_KSY[ksy_path] = import_spec(info)  # type: ignore
        return KaitaiParser(_PARSERS_BY_KSY[ksy_path])

    def parse(self, input_file_path_or_content: Union[str, Path, bytes, BytesIO]) -> KaitaiInspector:
        if isinstance(input_file_path_or_content, Path):
            input_file_path_or_content = str(input_file_path_or_content)
        if isinstance(input_file_path_or_content, str):
            struct = self.struct_type.from_file(input_file_path_or_content)
        elif isinstance(input_file_path_or_content, bytes):
            struct = self.struct_type.from_bytes(input_file_path_or_content)
        else:
            # Treat it like BytesIO
            struct = self.struct_type.from_io(input_file_path_or_content)
        struct._read()
        return KaitaiInspector(struct)



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class DimeMessage(KaitaiStruct):
    """Direct Internet Message Encapsulation (DIME)
    is an old Microsoft specification for sending and receiving
    SOAP messages along with additional attachments,
    like binary files, XML fragments, and even other
    SOAP messages, using standard transport protocols like HTTP.
    
    Sample file: `curl -LO
    https://github.com/kaitai-io/kaitai_struct_formats/files/5894723/scanner_withoptions.dump.gz
    && gunzip scanner_withoptions.dump.gz`
    
    .. seealso::
       Source - https://tools.ietf.org/html/draft-nielsen-dime-02
    
    
    .. seealso::
       Source - https://docs.microsoft.com/en-us/archive/msdn-magazine/2002/december/sending-files-attachments-and-soap-messages-via-dime
    
    
    .. seealso::
       Source - http://imrannazar.com/Parsing-the-DIME-Message-Format
    """

    class TypeFormats(Enum):
        unchanged = 0
        media_type = 1
        absolute_uri = 2
        unknown = 3
        none = 4
    SEQ_FIELDS = ["records"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['records']['start'] = self._io.pos()
        self.records = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['records']:
                self._debug['records']['arr'] = []
            self._debug['records']['arr'].append({'start': self._io.pos()})
            _t_records = DimeMessage.Record(self._io, self, self._root)
            _t_records._read()
            self.records.append(_t_records)
            self._debug['records']['arr'][len(self.records) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['records']['end'] = self._io.pos()

    class Padding(KaitaiStruct):
        """padding to the next 4-byte boundary."""
        SEQ_FIELDS = ["boundary_padding"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['boundary_padding']['start'] = self._io.pos()
            self.boundary_padding = self._io.read_bytes((-(self._io.pos()) % 4))
            self._debug['boundary_padding']['end'] = self._io.pos()


    class OptionField(KaitaiStruct):
        """the option field of the record."""
        SEQ_FIELDS = ["option_elements"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['option_elements']['start'] = self._io.pos()
            self.option_elements = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['option_elements']:
                    self._debug['option_elements']['arr'] = []
                self._debug['option_elements']['arr'].append({'start': self._io.pos()})
                _t_option_elements = DimeMessage.OptionElement(self._io, self, self._root)
                _t_option_elements._read()
                self.option_elements.append(_t_option_elements)
                self._debug['option_elements']['arr'][len(self.option_elements) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['option_elements']['end'] = self._io.pos()


    class OptionElement(KaitaiStruct):
        """one element of the option field."""
        SEQ_FIELDS = ["element_format", "len_element", "element_data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['element_format']['start'] = self._io.pos()
            self.element_format = self._io.read_u2be()
            self._debug['element_format']['end'] = self._io.pos()
            self._debug['len_element']['start'] = self._io.pos()
            self.len_element = self._io.read_u2be()
            self._debug['len_element']['end'] = self._io.pos()
            self._debug['element_data']['start'] = self._io.pos()
            self.element_data = self._io.read_bytes(self.len_element)
            self._debug['element_data']['end'] = self._io.pos()


    class Record(KaitaiStruct):
        """each individual fragment of the message."""
        SEQ_FIELDS = ["version", "is_first_record", "is_last_record", "is_chunk_record", "type_format", "reserved", "len_options", "len_id", "len_type", "len_data", "options", "options_padding", "id", "id_padding", "type", "type_padding", "data", "data_padding"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['version']['start'] = self._io.pos()
            self.version = self._io.read_bits_int_be(5)
            self._debug['version']['end'] = self._io.pos()
            self._debug['is_first_record']['start'] = self._io.pos()
            self.is_first_record = self._io.read_bits_int_be(1) != 0
            self._debug['is_first_record']['end'] = self._io.pos()
            self._debug['is_last_record']['start'] = self._io.pos()
            self.is_last_record = self._io.read_bits_int_be(1) != 0
            self._debug['is_last_record']['end'] = self._io.pos()
            self._debug['is_chunk_record']['start'] = self._io.pos()
            self.is_chunk_record = self._io.read_bits_int_be(1) != 0
            self._debug['is_chunk_record']['end'] = self._io.pos()
            self._debug['type_format']['start'] = self._io.pos()
            self.type_format = KaitaiStream.resolve_enum(DimeMessage.TypeFormats, self._io.read_bits_int_be(4))
            self._debug['type_format']['end'] = self._io.pos()
            self._debug['reserved']['start'] = self._io.pos()
            self.reserved = self._io.read_bits_int_be(4)
            self._debug['reserved']['end'] = self._io.pos()
            self._io.align_to_byte()
            self._debug['len_options']['start'] = self._io.pos()
            self.len_options = self._io.read_u2be()
            self._debug['len_options']['end'] = self._io.pos()
            self._debug['len_id']['start'] = self._io.pos()
            self.len_id = self._io.read_u2be()
            self._debug['len_id']['end'] = self._io.pos()
            self._debug['len_type']['start'] = self._io.pos()
            self.len_type = self._io.read_u2be()
            self._debug['len_type']['end'] = self._io.pos()
            self._debug['len_data']['start'] = self._io.pos()
            self.len_data = self._io.read_u4be()
            self._debug['len_data']['end'] = self._io.pos()
            self._debug['options']['start'] = self._io.pos()
            self._raw_options = self._io.read_bytes(self.len_options)
            _io__raw_options = KaitaiStream(BytesIO(self._raw_options))
            self.options = DimeMessage.OptionField(_io__raw_options, self, self._root)
            self.options._read()
            self._debug['options']['end'] = self._io.pos()
            self._debug['options_padding']['start'] = self._io.pos()
            self.options_padding = DimeMessage.Padding(self._io, self, self._root)
            self.options_padding._read()
            self._debug['options_padding']['end'] = self._io.pos()
            self._debug['id']['start'] = self._io.pos()
            self.id = (self._io.read_bytes(self.len_id)).decode(u"ASCII")
            self._debug['id']['end'] = self._io.pos()
            self._debug['id_padding']['start'] = self._io.pos()
            self.id_padding = DimeMessage.Padding(self._io, self, self._root)
            self.id_padding._read()
            self._debug['id_padding']['end'] = self._io.pos()
            self._debug['type']['start'] = self._io.pos()
            self.type = (self._io.read_bytes(self.len_type)).decode(u"ASCII")
            self._debug['type']['end'] = self._io.pos()
            self._debug['type_padding']['start'] = self._io.pos()
            self.type_padding = DimeMessage.Padding(self._io, self, self._root)
            self.type_padding._read()
            self._debug['type_padding']['end'] = self._io.pos()
            self._debug['data']['start'] = self._io.pos()
            self.data = self._io.read_bytes(self.len_data)
            self._debug['data']['end'] = self._io.pos()
            self._debug['data_padding']['start'] = self._io.pos()
            self.data_padding = DimeMessage.Padding(self._io, self, self._root)
            self.data_padding._read()
            self._debug['data_padding']['end'] = self._io.pos()






import base64
import codecs
from collections import defaultdict
import glob
import gzip
import json
import os
from xml.etree import ElementTree

from . import logger
from .fileutils import make_stream, FileStream
from .search import MultiSequenceSearch

log = logger.getStatusLogger("TRiD")

TRID_DEFS_URL = 'http://mark0.net/download/triddefs_xml.7z'

DEF_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "defs")

SERIALIZED_DEFS_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), "defs.json.gz")

SERIALIZED_FULL_TRIE = os.path.join(os.path.dirname(os.path.realpath(__file__)), "trie_full.gz")
SERIALIZED_PARTIAL_TRIE = os.path.join(os.path.dirname(os.path.realpath(__file__)), "trie_partial.gz")

DEFS = None

# These are the file formats that don't have to start at byte offset zero of the file:
CAN_BE_OFFSET = {'adobe_pdf', 'zip', 'html'}


class TRiDDef:
    def __init__(self, name, filetype, ext, mime, patterns, strings=()):
        self.name = name
        self.filetype = filetype
        self.ext = ext
        self.mime = mime
        self.patterns = patterns
        self.strings = strings
        self.can_be_offset = False
        for cbo in CAN_BE_OFFSET:
            if self.name.startswith(cbo):
                self.can_be_offset = True
                break

    @staticmethod
    def deserialize(json):
        patterns = [(ppos, base64.b64decode(pbytes)) for ppos, pbytes in json['patterns']]
        strings = tuple(base64.b64decode(s) for s in json['strings'])
        return TRiDDef(
            name=json['name'],
            filetype=json['filetype'],
            ext=json['ext'],
            mime=json['mime'],
            patterns=patterns,
            strings=strings
        )

    def serialize(self):
        return {
            'name': self.name,
            'filetype': self.filetype,
            'ext': self.ext,
            'mime': self.mime,
            'patterns': [[ppos, base64.b64encode(pbytes).decode('utf-8')] for ppos, pbytes in self.patterns],
            'strings': [base64.b64encode(s).decode('utf-8') for s in self.strings]
        }

    @staticmethod
    def load(xml_path):
        xml = ElementTree.parse(xml_path).getroot()
        filetype = xml.findtext("Info/FileType").strip()
        ext = xml.findtext("Info/Ext", default="").split("/")
        mime = xml.findtext("Info/Mime", default="").strip()
        patterns = []
        for pattern in xml.findall("FrontBlock/Pattern"):
            for data in pattern:
                ppos = 0
                if data.tag == "Pos":
                    ppos = int(data.text)
                elif data.tag == "Bytes":
                    if len(data.text) % 2:
                        data.text = '0' + data.text
                    pbytes = bytes.fromhex(data.text)
            patterns.append((ppos, pbytes))
        if xml.findtext("General/CheckStrings", default="False").strip() == "True":
            strings = tuple(string.text.replace("'", "\x00").encode('utf-8') for string in xml.findall("GlobalStrings/String"))
        else:
            strings = ()
        return TRiDDef(os.path.split(xml_path)[-1], filetype, ext, mime, patterns, strings)

    def match(self, file_stream, try_all_offsets=False):
        with make_stream(file_stream) as fs:
            if try_all_offsets and self.can_be_offset:
                offset_range = range(len(fs))
            else:
                offset_range = (0,)
            for global_offset in offset_range:
                for offset, pattern in self.patterns:
                    if global_offset + offset >= len(fs):
                        break
                    fs.seek(global_offset + offset)
                    if fs.read(len(pattern)) != pattern:
                        break
                else:
                    if FileStream(fs, start=global_offset).contains_all(*self.strings):
                        yield global_offset

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name!r}, filetype={self.filetype!r}, ext={self.ext!r}, mime={self.mime!r}, patterns={self.patterns!r}, strings={self.strings!r})"


class Matcher:
    def __init__(self, try_all_offsets=False, defs=None):
        load()
        if defs is None:
            defs = DEFS
            self.custom_defs = False
        else:
            self.custom_defs = True
        self.patterns = defaultdict(set)
        self.try_all_offsets = try_all_offsets
        log.status("Building multi-string search data structures...")
        for tdef in defs:
            if not try_all_offsets and not tdef.can_be_offset:
                continue
            log.status(f"Building multi-string search data structures... {tdef.name}")
            for pos, seq in tdef.patterns:
                self.patterns[seq].add((pos, tdef))
            for string in tdef.strings:
                self.patterns[string].add((None, tdef))
        if try_all_offsets:
            trie_path = SERIALIZED_FULL_TRIE
        else:
            trie_path = SERIALIZED_PARTIAL_TRIE
        # This is currently slower than just building the Trie from scratch every time!
        #if os.path.exists(trie_path):
        #    log.status("Loading the Cached Aho-Corasick Trie...")
        #   with gzip.open(trie_path, 'rb') as f:
        #        self.search = MultiSequenceSearch.load(f)
        #else:
        log.status("Constructing the Aho-Corasick Trie...")
        self.search = MultiSequenceSearch(*self.patterns.keys())
            # Commented out the caching because it is slower than just rebuilding the Trie every time:
            #log.status("Caching the Aho-Corasick Trie...")
            #with gzip.open(trie_path, 'wb') as f:
            #    self.search.save(f)
        log.clear_status()

    def match(self, file_stream, progress_callback=None):
        with make_stream(file_stream) as fs:
            if progress_callback is not None:
                fslen = len(fs)

                def callback(stream, pos):
                    progress_callback(pos, fslen)
                fs.add_listener(callback)
            if not self.try_all_offsets:
                # See if any of the TDEFs occur at offset zero
                prev_pos = fs.tell()
                for tdef in DEFS:
                    if not tdef.can_be_offset:
                        for offset in tdef.match(fs):
                            yield offset, tdef
                fs.seek(prev_pos)
            found_strings = defaultdict(set)
            partial_tdefs = set()
            yielded = defaultdict(set)
            for offset, source in self.search.search(fs):
                log.debug(f"Found byte sequence {source!r} at offset {offset} potentially matching "
                          f"{''.join(set(tdef.filetype for _, tdef in self.patterns[source]))}")
                found_strings[source].add(offset)
                # see if this constitutes the potential start of a tdef
                for expected_offset, tdef in self.patterns[source]:
                    if expected_offset is None \
                        or (tdef.can_be_offset and expected_offset <= offset) \
                            or (not tdef.can_be_offset and expected_offset == offset):
                        partial_tdefs.add(tdef)
                # see if this completes any partial tdefs
                for tdef in partial_tdefs:
                    first_pattern_offset, first_pattern_seq = tdef.patterns[0]
                    for match_pos in found_strings[first_pattern_seq]:
                        if (tdef.can_be_offset and first_pattern_offset > match_pos)\
                                or (not tdef.can_be_offset and first_pattern_offset != match_pos):
                            continue
                        potential_start = match_pos - first_pattern_offset
                        if potential_start in yielded[tdef]:
                            # We already found this match
                            continue
                        for pos, seq in tdef.patterns[1:]:
                            if pos + potential_start not in found_strings[seq]:
                                # We haven't found this other required pattern yet
                                break
                        else:
                            # Check that all of the strings have been found
                            if all(
                                any(
                                    pos >= potential_start for pos in found_strings[string]
                                ) for string in tdef.strings
                            ):
                                yield potential_start, tdef
                                yielded[tdef].add(potential_start)


def download_defs():
    import subprocess
    import urllib.request
    log.status('Downloading TRiD file definitions...')
    trid_defs_zip = os.path.join(os.path.dirname(os.path.realpath(__file__)), "triddefs_xml.7z")
    urllib.request.urlretrieve(TRID_DEFS_URL, trid_defs_zip)
    log.status('Extracting TRiD file definitions...')
    subprocess.check_call(['7za', 'x', trid_defs_zip], cwd=os.path.dirname(os.path.realpath(__file__)))
    log.clear_status()


def download_defs_if_necessary():
    if not os.path.exists(DEF_DIR):
        download_defs()


def build_defs_cache_if_necessary():
    if not os.path.exists(SERIALIZED_DEFS_PATH):
        load()


def load():
    global DEFS
    if DEFS is not None:
        return

    if not os.path.exists(SERIALIZED_DEFS_PATH):
        download_defs_if_necessary()

        log.status('Loading TRiD file definitions...')

        DEFS = []

        for xml_path in glob.glob(os.path.join(DEF_DIR, '**', '*.xml')):
            DEFS.append(TRiDDef.load(xml_path))
            log.status(f'Loading TRiD file definitions... {DEFS[-1].name}')

        log.status('Serializing TRiD definitions...')

        with gzip.open(SERIALIZED_DEFS_PATH, 'wb') as f:
            json.dump([d.serialize() for d in DEFS], codecs.getwriter('utf-8')(f))

        log.clear_status()
    else:
        log.status('Loading cached TRiD file definitions...')
        with gzip.open(SERIALIZED_DEFS_PATH, 'rb') as f:
            DEFS = [TRiDDef.deserialize(tdef) for tdef in json.load(codecs.getreader('utf-8')(f))]
        log.clear_status()



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

from polyfile.kaitai.parsers import php_serialized_value
class PharWithoutStub(KaitaiStruct):
    """A phar (PHP archive) file. The phar format is a custom archive format
    from the PHP ecosystem that is used to package a complete PHP library
    or application into a single self-contained archive.
    All phar archives start with an executable PHP stub, which can be used to
    allow executing or including phar files as if they were regular PHP scripts.
    PHP 5.3 and later include the phar extension, which adds native support for
    reading and manipulating phar files.
    
    The phar format was originally developed as part of the PEAR library
    PHP_Archive, first released in 2005. Later, a native PHP extension
    named "phar" was developed, which was first released on PECL in 2007,
    and is included with PHP 5.3 and later. The phar extension has effectively
    superseded the PHP_Archive library, which has not been updated since 2010.
    The phar extension is also no longer released independently on PECL;
    it is now developed and released as part of PHP itself.
    
    Because of current limitations in Kaitai Struct
    (seekaitai-io/kaitai_struct#158 and kaitai-io/kaitai_struct#538),
    the executable PHP stub that precedes the rest of the archive is not handled
    by this spec. Before parsing a phar using this spec, the stub must be
    removed manually.
    
    A phar's stub is terminated by the special token `__HALT_COMPILER();`
    (which may be followed by at most one space, the PHP tag end `?>`,
    and an optional line terminator). The stub termination sequence is
    immediately followed by the remaining parts of the phar format,
    as described in this spec.
    
    The phar stub usually contains code that loads the phar and runs
    a contained PHP file, but this is not required. A minimal valid phar stub
    is `<?php __HALT_COMPILER();` - such a stub makes it impossible to execute
    the phar directly, but still allows loading or manipulating it using the
    phar extension.
    
    Note: The phar format does not specify any encoding for text fields
    (stub, alias name, and all file names), so these fields may contain arbitrary
    binary data. The actual text encoding used in a specific phar file usually
    depends on the application that created the phar, and on the
    standard encoding of the system on which the phar was created.
    
    .. seealso::
       Source - https://www.php.net/manual/en/phar.fileformat.php
    
    
    .. seealso::
       Source - https://github.com/php/php-src/tree/master/ext/phar
    
    
    .. seealso::
       Source - https://svn.php.net/viewvc/pecl/phar/
    
    
    .. seealso::
       Source - https://svn.php.net/viewvc/pear/packages/PHP_Archive/
    """

    class SignatureType(Enum):
        md5 = 1
        sha1 = 2
        sha256 = 4
        sha512 = 8
        openssl = 16
    SEQ_FIELDS = ["manifest", "files", "signature"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['manifest']['start'] = self._io.pos()
        self.manifest = PharWithoutStub.Manifest(self._io, self, self._root)
        self.manifest._read()
        self._debug['manifest']['end'] = self._io.pos()
        self._debug['files']['start'] = self._io.pos()
        self.files = [None] * (self.manifest.num_files)
        for i in range(self.manifest.num_files):
            if not 'arr' in self._debug['files']:
                self._debug['files']['arr'] = []
            self._debug['files']['arr'].append({'start': self._io.pos()})
            self.files[i] = self._io.read_bytes(self.manifest.file_entries[i].len_data_compressed)
            self._debug['files']['arr'][i]['end'] = self._io.pos()

        self._debug['files']['end'] = self._io.pos()
        if self.manifest.flags.has_signature:
            self._debug['signature']['start'] = self._io.pos()
            self._raw_signature = self._io.read_bytes_full()
            _io__raw_signature = KaitaiStream(BytesIO(self._raw_signature))
            self.signature = PharWithoutStub.Signature(_io__raw_signature, self, self._root)
            self.signature._read()
            self._debug['signature']['end'] = self._io.pos()


    class SerializedValue(KaitaiStruct):
        SEQ_FIELDS = ["raw"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['raw']['start'] = self._io.pos()
            self.raw = self._io.read_bytes_full()
            self._debug['raw']['end'] = self._io.pos()

        @property
        def parsed(self):
            """The serialized value, parsed as a structure."""
            if hasattr(self, '_m_parsed'):
                return self._m_parsed if hasattr(self, '_m_parsed') else None

            _pos = self._io.pos()
            self._io.seek(0)
            self._debug['_m_parsed']['start'] = self._io.pos()
            self._m_parsed = php_serialized_value.PhpSerializedValue(self._io)
            self._m_parsed._read()
            self._debug['_m_parsed']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_parsed if hasattr(self, '_m_parsed') else None


    class Signature(KaitaiStruct):
        SEQ_FIELDS = ["data", "type", "magic"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['data']['start'] = self._io.pos()
            self.data = self._io.read_bytes(((self._io.size() - self._io.pos()) - 8))
            self._debug['data']['end'] = self._io.pos()
            self._debug['type']['start'] = self._io.pos()
            self.type = KaitaiStream.resolve_enum(PharWithoutStub.SignatureType, self._io.read_u4le())
            self._debug['type']['end'] = self._io.pos()
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x47\x42\x4D\x42":
                raise kaitaistruct.ValidationNotEqualError(b"\x47\x42\x4D\x42", self.magic, self._io, u"/types/signature/seq/2")


    class FileFlags(KaitaiStruct):
        SEQ_FIELDS = ["value"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['value']['start'] = self._io.pos()
            self.value = self._io.read_u4le()
            self._debug['value']['end'] = self._io.pos()

        @property
        def permissions(self):
            """The file's permission bits."""
            if hasattr(self, '_m_permissions'):
                return self._m_permissions if hasattr(self, '_m_permissions') else None

            self._m_permissions = (self.value & 511)
            return self._m_permissions if hasattr(self, '_m_permissions') else None

        @property
        def zlib_compressed(self):
            """Whether this file's data is stored using zlib compression."""
            if hasattr(self, '_m_zlib_compressed'):
                return self._m_zlib_compressed if hasattr(self, '_m_zlib_compressed') else None

            self._m_zlib_compressed = (self.value & 4096) != 0
            return self._m_zlib_compressed if hasattr(self, '_m_zlib_compressed') else None

        @property
        def bzip2_compressed(self):
            """Whether this file's data is stored using bzip2 compression."""
            if hasattr(self, '_m_bzip2_compressed'):
                return self._m_bzip2_compressed if hasattr(self, '_m_bzip2_compressed') else None

            self._m_bzip2_compressed = (self.value & 8192) != 0
            return self._m_bzip2_compressed if hasattr(self, '_m_bzip2_compressed') else None


    class ApiVersion(KaitaiStruct):
        """A phar API version number. This version number is meant to indicate
        which features are used in a specific phar, so that tools reading
        the phar can easily check that they support all necessary features.
        
        The following API versions exist so far:
        
        * 0.5, 0.6, 0.7, 0.7.1: The first official API versions. At this point,
          the phar format was only used by the PHP_Archive library, and the
          API version numbers were identical to the PHP_Archive versions that
          supported them. Development of the native phar extension started around
          API version 0.7. These API versions could only be queried using the
          `PHP_Archive::APIversion()` method, but were not stored physically
          in archives. These API versions are not supported by this spec.
        * 0.8.0: Used by PHP_Archive 0.8.0 (released 2006-07-18) and
          later development versions of the phar extension. This is the first
          version number to be physically stored in archives. This API version
          is not supported by this spec.
        * 0.9.0: Used by later development/early beta versions of the
          phar extension. Also temporarily used by PHP_Archive 0.9.0
          (released 2006-12-15), but reverted back to API version 0.8.0 in
          PHP_Archive 0.9.1 (released 2007-01-05).
        * 1.0.0: Supported since PHP_Archive 0.10.0 (released 2007-05-29)
          and phar extension 1.0.0 (released 2007-03-28). This is the first
          stable, forwards-compatible and documented version of the format.
        * 1.1.0: Supported since PHP_Archive 0.12.0 (released 2015-07-06)
          and phar extension 1.1.0 (released 2007-04-12). Adds SHA-256 and
          SHA-512 signature types.
        * 1.1.1: Supported since phar extension 2.0.0 (released 2009-07-29 and
          included with PHP 5.3 and later). (PHP_Archive 0.12.0 also supports
          all features from API verison 1.1.1, but it reports API version 1.1.0.)
          Adds the OpenSSL signature type and support for storing
          empty directories.
        """
        SEQ_FIELDS = ["release", "major", "minor", "unused"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['release']['start'] = self._io.pos()
            self.release = self._io.read_bits_int_be(4)
            self._debug['release']['end'] = self._io.pos()
            self._debug['major']['start'] = self._io.pos()
            self.major = self._io.read_bits_int_be(4)
            self._debug['major']['end'] = self._io.pos()
            self._debug['minor']['start'] = self._io.pos()
            self.minor = self._io.read_bits_int_be(4)
            self._debug['minor']['end'] = self._io.pos()
            self._debug['unused']['start'] = self._io.pos()
            self.unused = self._io.read_bits_int_be(4)
            self._debug['unused']['end'] = self._io.pos()


    class GlobalFlags(KaitaiStruct):
        SEQ_FIELDS = ["value"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['value']['start'] = self._io.pos()
            self.value = self._io.read_u4le()
            self._debug['value']['end'] = self._io.pos()

        @property
        def any_zlib_compressed(self):
            """Whether any of the files in this phar are stored using
            zlib compression.
            """
            if hasattr(self, '_m_any_zlib_compressed'):
                return self._m_any_zlib_compressed if hasattr(self, '_m_any_zlib_compressed') else None

            self._m_any_zlib_compressed = (self.value & 4096) != 0
            return self._m_any_zlib_compressed if hasattr(self, '_m_any_zlib_compressed') else None

        @property
        def any_bzip2_compressed(self):
            """Whether any of the files in this phar are stored using
            bzip2 compression.
            """
            if hasattr(self, '_m_any_bzip2_compressed'):
                return self._m_any_bzip2_compressed if hasattr(self, '_m_any_bzip2_compressed') else None

            self._m_any_bzip2_compressed = (self.value & 8192) != 0
            return self._m_any_bzip2_compressed if hasattr(self, '_m_any_bzip2_compressed') else None

        @property
        def has_signature(self):
            """Whether this phar contains a signature."""
            if hasattr(self, '_m_has_signature'):
                return self._m_has_signature if hasattr(self, '_m_has_signature') else None

            self._m_has_signature = (self.value & 65536) != 0
            return self._m_has_signature if hasattr(self, '_m_has_signature') else None


    class Manifest(KaitaiStruct):
        SEQ_FIELDS = ["len_manifest", "num_files", "api_version", "flags", "len_alias", "alias", "len_metadata", "metadata", "file_entries"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_manifest']['start'] = self._io.pos()
            self.len_manifest = self._io.read_u4le()
            self._debug['len_manifest']['end'] = self._io.pos()
            self._debug['num_files']['start'] = self._io.pos()
            self.num_files = self._io.read_u4le()
            self._debug['num_files']['end'] = self._io.pos()
            self._debug['api_version']['start'] = self._io.pos()
            self.api_version = PharWithoutStub.ApiVersion(self._io, self, self._root)
            self.api_version._read()
            self._debug['api_version']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = PharWithoutStub.GlobalFlags(self._io, self, self._root)
            self.flags._read()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['len_alias']['start'] = self._io.pos()
            self.len_alias = self._io.read_u4le()
            self._debug['len_alias']['end'] = self._io.pos()
            self._debug['alias']['start'] = self._io.pos()
            self.alias = self._io.read_bytes(self.len_alias)
            self._debug['alias']['end'] = self._io.pos()
            self._debug['len_metadata']['start'] = self._io.pos()
            self.len_metadata = self._io.read_u4le()
            self._debug['len_metadata']['end'] = self._io.pos()
            if self.len_metadata != 0:
                self._debug['metadata']['start'] = self._io.pos()
                self._raw_metadata = self._io.read_bytes(self.len_metadata)
                _io__raw_metadata = KaitaiStream(BytesIO(self._raw_metadata))
                self.metadata = PharWithoutStub.SerializedValue(_io__raw_metadata, self, self._root)
                self.metadata._read()
                self._debug['metadata']['end'] = self._io.pos()

            self._debug['file_entries']['start'] = self._io.pos()
            self.file_entries = [None] * (self.num_files)
            for i in range(self.num_files):
                if not 'arr' in self._debug['file_entries']:
                    self._debug['file_entries']['arr'] = []
                self._debug['file_entries']['arr'].append({'start': self._io.pos()})
                _t_file_entries = PharWithoutStub.FileEntry(self._io, self, self._root)
                _t_file_entries._read()
                self.file_entries[i] = _t_file_entries
                self._debug['file_entries']['arr'][i]['end'] = self._io.pos()

            self._debug['file_entries']['end'] = self._io.pos()


    class FileEntry(KaitaiStruct):
        SEQ_FIELDS = ["len_filename", "filename", "len_data_uncompressed", "timestamp", "len_data_compressed", "crc32", "flags", "len_metadata", "metadata"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_filename']['start'] = self._io.pos()
            self.len_filename = self._io.read_u4le()
            self._debug['len_filename']['end'] = self._io.pos()
            self._debug['filename']['start'] = self._io.pos()
            self.filename = self._io.read_bytes(self.len_filename)
            self._debug['filename']['end'] = self._io.pos()
            self._debug['len_data_uncompressed']['start'] = self._io.pos()
            self.len_data_uncompressed = self._io.read_u4le()
            self._debug['len_data_uncompressed']['end'] = self._io.pos()
            self._debug['timestamp']['start'] = self._io.pos()
            self.timestamp = self._io.read_u4le()
            self._debug['timestamp']['end'] = self._io.pos()
            self._debug['len_data_compressed']['start'] = self._io.pos()
            self.len_data_compressed = self._io.read_u4le()
            self._debug['len_data_compressed']['end'] = self._io.pos()
            self._debug['crc32']['start'] = self._io.pos()
            self.crc32 = self._io.read_u4le()
            self._debug['crc32']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = PharWithoutStub.FileFlags(self._io, self, self._root)
            self.flags._read()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['len_metadata']['start'] = self._io.pos()
            self.len_metadata = self._io.read_u4le()
            self._debug['len_metadata']['end'] = self._io.pos()
            if self.len_metadata != 0:
                self._debug['metadata']['start'] = self._io.pos()
                self._raw_metadata = self._io.read_bytes(self.len_metadata)
                _io__raw_metadata = KaitaiStream(BytesIO(self._raw_metadata))
                self.metadata = PharWithoutStub.SerializedValue(_io__raw_metadata, self, self._root)
                self.metadata._read()
                self._debug['metadata']['end'] = self._io.pos()







# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections
from enum import Enum


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Ines(KaitaiStruct):
    """
    .. seealso::
       Source - https://wiki.nesdev.com/w/index.php/INES
    """
    SEQ_FIELDS = ["header", "trainer", "prg_rom", "chr_rom", "playchoice10", "title"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['header']['start'] = self._io.pos()
        self._raw_header = self._io.read_bytes(16)
        _io__raw_header = KaitaiStream(BytesIO(self._raw_header))
        self.header = Ines.Header(_io__raw_header, self, self._root)
        self.header._read()
        self._debug['header']['end'] = self._io.pos()
        if self.header.f6.trainer:
            self._debug['trainer']['start'] = self._io.pos()
            self.trainer = self._io.read_bytes(512)
            self._debug['trainer']['end'] = self._io.pos()

        self._debug['prg_rom']['start'] = self._io.pos()
        self.prg_rom = self._io.read_bytes((self.header.len_prg_rom * 16384))
        self._debug['prg_rom']['end'] = self._io.pos()
        self._debug['chr_rom']['start'] = self._io.pos()
        self.chr_rom = self._io.read_bytes((self.header.len_chr_rom * 8192))
        self._debug['chr_rom']['end'] = self._io.pos()
        if self.header.f7.playchoice10:
            self._debug['playchoice10']['start'] = self._io.pos()
            self.playchoice10 = Ines.Playchoice10(self._io, self, self._root)
            self.playchoice10._read()
            self._debug['playchoice10']['end'] = self._io.pos()

        if not (self._io.is_eof()):
            self._debug['title']['start'] = self._io.pos()
            self.title = (self._io.read_bytes_full()).decode(u"ASCII")
            self._debug['title']['end'] = self._io.pos()


    class Header(KaitaiStruct):
        SEQ_FIELDS = ["magic", "len_prg_rom", "len_chr_rom", "f6", "f7", "len_prg_ram", "f9", "f10", "reserved"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x4E\x45\x53\x1A":
                raise kaitaistruct.ValidationNotEqualError(b"\x4E\x45\x53\x1A", self.magic, self._io, u"/types/header/seq/0")
            self._debug['len_prg_rom']['start'] = self._io.pos()
            self.len_prg_rom = self._io.read_u1()
            self._debug['len_prg_rom']['end'] = self._io.pos()
            self._debug['len_chr_rom']['start'] = self._io.pos()
            self.len_chr_rom = self._io.read_u1()
            self._debug['len_chr_rom']['end'] = self._io.pos()
            self._debug['f6']['start'] = self._io.pos()
            self._raw_f6 = self._io.read_bytes(1)
            _io__raw_f6 = KaitaiStream(BytesIO(self._raw_f6))
            self.f6 = Ines.Header.F6(_io__raw_f6, self, self._root)
            self.f6._read()
            self._debug['f6']['end'] = self._io.pos()
            self._debug['f7']['start'] = self._io.pos()
            self._raw_f7 = self._io.read_bytes(1)
            _io__raw_f7 = KaitaiStream(BytesIO(self._raw_f7))
            self.f7 = Ines.Header.F7(_io__raw_f7, self, self._root)
            self.f7._read()
            self._debug['f7']['end'] = self._io.pos()
            self._debug['len_prg_ram']['start'] = self._io.pos()
            self.len_prg_ram = self._io.read_u1()
            self._debug['len_prg_ram']['end'] = self._io.pos()
            self._debug['f9']['start'] = self._io.pos()
            self._raw_f9 = self._io.read_bytes(1)
            _io__raw_f9 = KaitaiStream(BytesIO(self._raw_f9))
            self.f9 = Ines.Header.F9(_io__raw_f9, self, self._root)
            self.f9._read()
            self._debug['f9']['end'] = self._io.pos()
            self._debug['f10']['start'] = self._io.pos()
            self._raw_f10 = self._io.read_bytes(1)
            _io__raw_f10 = KaitaiStream(BytesIO(self._raw_f10))
            self.f10 = Ines.Header.F10(_io__raw_f10, self, self._root)
            self.f10._read()
            self._debug['f10']['end'] = self._io.pos()
            self._debug['reserved']['start'] = self._io.pos()
            self.reserved = self._io.read_bytes(5)
            self._debug['reserved']['end'] = self._io.pos()
            if not self.reserved == b"\x00\x00\x00\x00\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x00\x00\x00\x00\x00", self.reserved, self._io, u"/types/header/seq/8")

        class F6(KaitaiStruct):
            """
            .. seealso::
               Source - https://wiki.nesdev.com/w/index.php/INES#Flags_6
            """

            class Mirroring(Enum):
                horizontal = 0
                vertical = 1
            SEQ_FIELDS = ["lower_mapper", "four_screen", "trainer", "has_battery_ram", "mirroring"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['lower_mapper']['start'] = self._io.pos()
                self.lower_mapper = self._io.read_bits_int_be(4)
                self._debug['lower_mapper']['end'] = self._io.pos()
                self._debug['four_screen']['start'] = self._io.pos()
                self.four_screen = self._io.read_bits_int_be(1) != 0
                self._debug['four_screen']['end'] = self._io.pos()
                self._debug['trainer']['start'] = self._io.pos()
                self.trainer = self._io.read_bits_int_be(1) != 0
                self._debug['trainer']['end'] = self._io.pos()
                self._debug['has_battery_ram']['start'] = self._io.pos()
                self.has_battery_ram = self._io.read_bits_int_be(1) != 0
                self._debug['has_battery_ram']['end'] = self._io.pos()
                self._debug['mirroring']['start'] = self._io.pos()
                self.mirroring = KaitaiStream.resolve_enum(Ines.Header.F6.Mirroring, self._io.read_bits_int_be(1))
                self._debug['mirroring']['end'] = self._io.pos()


        class F7(KaitaiStruct):
            """
            .. seealso::
               Source - https://wiki.nesdev.com/w/index.php/INES#Flags_7
            """
            SEQ_FIELDS = ["upper_mapper", "format", "playchoice10", "vs_unisystem"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['upper_mapper']['start'] = self._io.pos()
                self.upper_mapper = self._io.read_bits_int_be(4)
                self._debug['upper_mapper']['end'] = self._io.pos()
                self._debug['format']['start'] = self._io.pos()
                self.format = self._io.read_bits_int_be(2)
                self._debug['format']['end'] = self._io.pos()
                self._debug['playchoice10']['start'] = self._io.pos()
                self.playchoice10 = self._io.read_bits_int_be(1) != 0
                self._debug['playchoice10']['end'] = self._io.pos()
                self._debug['vs_unisystem']['start'] = self._io.pos()
                self.vs_unisystem = self._io.read_bits_int_be(1) != 0
                self._debug['vs_unisystem']['end'] = self._io.pos()


        class F9(KaitaiStruct):
            """
            .. seealso::
               Source - https://wiki.nesdev.com/w/index.php/INES#Flags_9
            """

            class TvSystem(Enum):
                ntsc = 0
                pal = 1
            SEQ_FIELDS = ["reserved", "tv_system"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved']['start'] = self._io.pos()
                self.reserved = self._io.read_bits_int_be(7)
                self._debug['reserved']['end'] = self._io.pos()
                self._debug['tv_system']['start'] = self._io.pos()
                self.tv_system = KaitaiStream.resolve_enum(Ines.Header.F9.TvSystem, self._io.read_bits_int_be(1))
                self._debug['tv_system']['end'] = self._io.pos()


        class F10(KaitaiStruct):
            """
            .. seealso::
               Source - https://wiki.nesdev.com/w/index.php/INES#Flags_10
            """

            class TvSystem(Enum):
                ntsc = 0
                dual1 = 1
                pal = 2
                dual2 = 3
            SEQ_FIELDS = ["reserved1", "bus_conflict", "prg_ram", "reserved2", "tv_system"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved1']['start'] = self._io.pos()
                self.reserved1 = self._io.read_bits_int_be(2)
                self._debug['reserved1']['end'] = self._io.pos()
                self._debug['bus_conflict']['start'] = self._io.pos()
                self.bus_conflict = self._io.read_bits_int_be(1) != 0
                self._debug['bus_conflict']['end'] = self._io.pos()
                self._debug['prg_ram']['start'] = self._io.pos()
                self.prg_ram = self._io.read_bits_int_be(1) != 0
                self._debug['prg_ram']['end'] = self._io.pos()
                self._debug['reserved2']['start'] = self._io.pos()
                self.reserved2 = self._io.read_bits_int_be(2)
                self._debug['reserved2']['end'] = self._io.pos()
                self._debug['tv_system']['start'] = self._io.pos()
                self.tv_system = KaitaiStream.resolve_enum(Ines.Header.F10.TvSystem, self._io.read_bits_int_be(2))
                self._debug['tv_system']['end'] = self._io.pos()


        @property
        def mapper(self):
            """
            .. seealso::
               Source - https://wiki.nesdev.com/w/index.php/Mapper
            """
            if hasattr(self, '_m_mapper'):
                return self._m_mapper if hasattr(self, '_m_mapper') else None

            self._m_mapper = (self.f6.lower_mapper | (self.f7.upper_mapper << 4))
            return self._m_mapper if hasattr(self, '_m_mapper') else None


    class Playchoice10(KaitaiStruct):
        """
        .. seealso::
           Source - http://wiki.nesdev.com/w/index.php/PC10_ROM-Images
        """
        SEQ_FIELDS = ["inst_rom", "prom"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['inst_rom']['start'] = self._io.pos()
            self.inst_rom = self._io.read_bytes(8192)
            self._debug['inst_rom']['end'] = self._io.pos()
            self._debug['prom']['start'] = self._io.pos()
            self.prom = Ines.Playchoice10.Prom(self._io, self, self._root)
            self.prom._read()
            self._debug['prom']['end'] = self._io.pos()

        class Prom(KaitaiStruct):
            SEQ_FIELDS = ["data", "counter_out"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['data']['start'] = self._io.pos()
                self.data = self._io.read_bytes(16)
                self._debug['data']['end'] = self._io.pos()
                self._debug['counter_out']['start'] = self._io.pos()
                self.counter_out = self._io.read_bytes(16)
                self._debug['counter_out']['end'] = self._io.pos()







# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Uimage(KaitaiStruct):
    """The new uImage format allows more flexibility in handling images of various
    types (kernel, ramdisk, etc.), it also enhances integrity protection of images
    with sha1 and md5 checksums.
    
    .. seealso::
       Source - https://source.denx.de/u-boot/u-boot/-/raw/e4dba4b/include/image.h
    """

    class UimageOs(Enum):
        invalid = 0
        openbsd = 1
        netbsd = 2
        freebsd = 3
        bsd4_4 = 4
        linux = 5
        svr4 = 6
        esix = 7
        solaris = 8
        irix = 9
        sco = 10
        dell = 11
        ncr = 12
        lynxos = 13
        vxworks = 14
        psos = 15
        qnx = 16
        u_boot = 17
        rtems = 18
        artos = 19
        unity = 20
        integrity = 21
        ose = 22
        plan9 = 23
        openrtos = 24
        arm_trusted_firmware = 25
        tee = 26
        opensbi = 27
        efi = 28

    class UimageArch(Enum):
        invalid = 0
        alpha = 1
        arm = 2
        i386 = 3
        ia64 = 4
        mips = 5
        mips64 = 6
        ppc = 7
        s390 = 8
        sh = 9
        sparc = 10
        sparc64 = 11
        m68k = 12
        nios = 13
        microblaze = 14
        nios2 = 15
        blackfin = 16
        avr32 = 17
        st200 = 18
        sandbox = 19
        nds32 = 20
        openrisc = 21
        arm64 = 22
        arc = 23
        x86_64 = 24
        xtensa = 25
        riscv = 26

    class UimageComp(Enum):
        none = 0
        gzip = 1
        bzip2 = 2
        lzma = 3
        lzo = 4
        lz4 = 5
        zstd = 6

    class UimageType(Enum):
        invalid = 0
        standalone = 1
        kernel = 2
        ramdisk = 3
        multi = 4
        firmware = 5
        script = 6
        filesystem = 7
        flatdt = 8
        kwbimage = 9
        imximage = 10
        ublimage = 11
        omapimage = 12
        aisimage = 13
        kernel_noload = 14
        pblimage = 15
        mxsimage = 16
        gpimage = 17
        atmelimage = 18
        socfpgaimage = 19
        x86_setup = 20
        lpc32xximage = 21
        loadable = 22
        rkimage = 23
        rksd = 24
        rkspi = 25
        zynqimage = 26
        zynqmpimage = 27
        zynqmpbif = 28
        fpga = 29
        vybridimage = 30
        tee = 31
        firmware_ivt = 32
        pmmc = 33
        stm32image = 34
        socfpgaimage_v1 = 35
        mtkimage = 36
        imx8mimage = 37
        imx8image = 38
        copro = 39
        sunxi_egon = 40
    SEQ_FIELDS = ["header", "data"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['header']['start'] = self._io.pos()
        self.header = Uimage.Uheader(self._io, self, self._root)
        self.header._read()
        self._debug['header']['end'] = self._io.pos()
        self._debug['data']['start'] = self._io.pos()
        self.data = self._io.read_bytes(self.header.len_image)
        self._debug['data']['end'] = self._io.pos()

    class Uheader(KaitaiStruct):
        SEQ_FIELDS = ["magic", "header_crc", "timestamp", "len_image", "load_address", "entry_address", "data_crc", "os_type", "architecture", "image_type", "compression_type", "name"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x27\x05\x19\x56":
                raise kaitaistruct.ValidationNotEqualError(b"\x27\x05\x19\x56", self.magic, self._io, u"/types/uheader/seq/0")
            self._debug['header_crc']['start'] = self._io.pos()
            self.header_crc = self._io.read_u4be()
            self._debug['header_crc']['end'] = self._io.pos()
            self._debug['timestamp']['start'] = self._io.pos()
            self.timestamp = self._io.read_u4be()
            self._debug['timestamp']['end'] = self._io.pos()
            self._debug['len_image']['start'] = self._io.pos()
            self.len_image = self._io.read_u4be()
            self._debug['len_image']['end'] = self._io.pos()
            self._debug['load_address']['start'] = self._io.pos()
            self.load_address = self._io.read_u4be()
            self._debug['load_address']['end'] = self._io.pos()
            self._debug['entry_address']['start'] = self._io.pos()
            self.entry_address = self._io.read_u4be()
            self._debug['entry_address']['end'] = self._io.pos()
            self._debug['data_crc']['start'] = self._io.pos()
            self.data_crc = self._io.read_u4be()
            self._debug['data_crc']['end'] = self._io.pos()
            self._debug['os_type']['start'] = self._io.pos()
            self.os_type = KaitaiStream.resolve_enum(Uimage.UimageOs, self._io.read_u1())
            self._debug['os_type']['end'] = self._io.pos()
            self._debug['architecture']['start'] = self._io.pos()
            self.architecture = KaitaiStream.resolve_enum(Uimage.UimageArch, self._io.read_u1())
            self._debug['architecture']['end'] = self._io.pos()
            self._debug['image_type']['start'] = self._io.pos()
            self.image_type = KaitaiStream.resolve_enum(Uimage.UimageType, self._io.read_u1())
            self._debug['image_type']['end'] = self._io.pos()
            self._debug['compression_type']['start'] = self._io.pos()
            self.compression_type = KaitaiStream.resolve_enum(Uimage.UimageComp, self._io.read_u1())
            self._debug['compression_type']['end'] = self._io.pos()
            self._debug['name']['start'] = self._io.pos()
            self.name = (KaitaiStream.bytes_terminate(self._io.read_bytes(32), 0, False)).decode(u"UTF-8")
            self._debug['name']['end'] = self._io.pos()



